# Standard library imports.
from cStringIO import StringIO
import sys

# Enthought library imports.
from traits.api import Either, Enum, HasStrictTraits, Int, Instance, List, Str


class LayoutItem(HasStrictTraits):
    """ The base class for all Task-related layout objects.
    """

    def __repr__(self):
        return self.pformat()

    def iterleaves(self):
        yield self

    def pargs(self):
        return []

    def pformat(self, indent=0, multiline=False):
        """ Pretty-format the layout item. Returns a string.
        """
        stream = StringIO()
        self.pstream(stream, indent, multiline)
        return stream.getvalue()

    def pprint(self, indent=0, multiline=False):
        """ Pretty-prints the layout item.
        """
        self.pstream(sys.stdout, indent, multiline)

    def pstream(self, stream, indent=0, multiline=False):
        """ Pretty-formats the layout item to a stream.
        """
        call = self.__class__.__name__ + '('
        indent += len(call)
        stream.write(call)

        args = [(None, arg) for arg in self.pargs()]
        traits = []
        for name, trait in sorted(self.traits().iteritems()):
            if not trait.pretty_skip and not trait.transient:
                value = getattr(self, name)
                if trait.default != value:
                    traits.append((name, value))
        traits.sort()
        args.extend(traits)

        for i, (name, value) in enumerate(args):
            arg_indent = indent
            if name:
                arg_indent += len(name) + 1
                stream.write(name + '=')
            if isinstance(value, LayoutItem):
                value.pstream(stream, arg_indent, multiline)
            else:
                stream.write(repr(value))
            if i < len(args) - 1:
                stream.write(',')
                if multiline:
                    stream.write('\n' + indent * ' ')
                else:
                    stream.write(' ')

        stream.write(')')


class LayoutContainer(LayoutItem):
    """ The base class for all layout items that contain other layout items.
    """

    items = List(pretty_skip=True)

    def __init__(self, *items, **traits):
        # Items may either be specified as a positional arg or a kwarg.
        if items:
            if 'items' in traits:
                raise ValueError(
                    "Received 'items' as positional and keyword argument."
                )
            else:
                traits['items'] = list(items)
        super(LayoutContainer, self).__init__(**traits)

    def iterleaves(self):
        for item in self.items:
            for leaf in item.iterleaves():
                yield leaf

    def pargs(self):
        return self.items


class PaneItem(LayoutItem):
    """ A pane in a Task layout.
    """

    # The ID of the item. If the item refers to a TaskPane, this is the ID of
    # that TaskPane.
    id = Either(Str, Int, default='', pretty_skip=True)

    # The width of the pane in pixels. If not specified, the pane will be sized
    # according to its size hint.
    width = Int(-1)

    # The height of the pane in pixels. If not specified, the pane will be
    # sized according to its size hint.
    height = Int(-1)

    def __init__(self, id='', **traits):
        super(PaneItem, self).__init__(**traits)
        self.id = id

    def pargs(self):
        return [self.id]


class Tabbed(LayoutContainer):
    """ A tab area in a Task layout.
    """

    # A tabbed layout can only contain PaneItems as sub-items. Splitters and
    # other Tabbed layouts are not allowed.
    items = List(PaneItem, pretty_skip=True)

    # The ID of the TaskPane which is active in layout. If not specified, the
    # first pane is active.
    active_tab = Either(Str, Int, default='')


class Splitter(LayoutContainer):
    """ A split area in a Task layout.
    """

    # The orientation of the splitter.
    orientation = Enum('horizontal', 'vertical')

    # The sub-items of the splitter, which are PaneItems, Tabbed layouts, and
    # other Splitters.
    items = List(Either(
        PaneItem,
        Tabbed,
        Instance('pyface.tasks.task_layout.Splitter')), pretty_skip=True)


class HSplitter(Splitter):
    """ A convenience class for horizontal splitters.
    """
    orientation = Str('horizontal')


class VSplitter(Splitter):
    """ A convenience class for vertical splitters.
    """
    orientation = Str('vertical')


class DockLayout(LayoutItem):
    """ The layout for a main window's dock area.
    """

    # The layouts for the task's dock panes.
    left = Either(PaneItem, Tabbed, Splitter)
    right = Either(PaneItem, Tabbed, Splitter)
    top = Either(PaneItem, Tabbed, Splitter)
    bottom = Either(PaneItem, Tabbed, Splitter)

    # Assignments of dock areas to the window's corners. By default, the top
    # and bottom dock areas extend into both of the top and both of the bottom
    # corners, respectively.
    top_left_corner = Enum('top', 'left')
    top_right_corner = Enum('top', 'right')
    bottom_left_corner = Enum('bottom', 'left')
    bottom_right_corner = Enum('bottom', 'right')


class TaskLayout(DockLayout):
    """ The layout for a Task.
    """

    # The ID of the task for which this is a layout.
    id = Str



""" The help doc implementation.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!

from apptools.preferences.api import PreferencesHelper
from traits.api import Either, File, Str, provides, Bool

from i_help_doc import IHelpDoc

@provides(IHelpDoc)
class HelpDoc(PreferencesHelper):
    """ The implementation for help docs.

    A help doc is defined by a UI label, a filename, and a viewer program.
    """


    #### IHelpDoc interface / Preferences ######################################

    # NOTE: This class inherits preferences_path from PreferencesHelper.

    # The UI label for the help doc, which appears in menus or dialogs.
    label = Str

    # The path to the document, which can be full, or relative to the Python
    # installation directory (sys.prefix).
    filename = File

    # Is this a url?
    url = Bool(False)

    # The program to use to view the document. 'browser' means the platform
    # default web browser. Otherwise, it is a command to run, which may be
    # in the program search path of the current environment, or an absolute
    # path to a program.
    viewer = Either('browser', Str)




""" The nodes used in the service registry browser tree. """


# Enthought library imports.
from envisage.api import IExtensionPoint, IServiceRegistry
from traits.api import Any, Dict, HasTraits, Instance, Int, List
from traits.api import Property, Str, Undefined
from traitsui.api import TreeNode

# fixme: non-api imports.
from traitsui.value_tree import SingleValueTreeNodeObject
from traitsui.value_tree import value_tree_nodes



class ServiceModel(HasTraits):
    """ A model of a registered service. """

    # The service Id.
    id = Int

    # The service object.
    obj = Any

    # The service properties.
    properties = Dict


class ProtocolModel(HasTraits):
    """ A model of all of the services of a single protocol. """

    # The protocol (i.e. type) name.
    name = Str

    # All of the services offered for the protocol.
    services = List(ServiceModel)


class ServiceRegistryModel(HasTraits):
    """ A model of a service registry.

    This model groups services by protocol.

    """

    # The service registry that we are a model of.
    service_registry = Instance(IServiceRegistry)

    # The protocols contained in the registry.
    protocols = Property(List(ProtocolModel))

    ###########################################################################
    # 'ServiceRegistryModel' interface.
    ###########################################################################

    def _get_protocols(self):
        """ Trait property getter. """

        # fixme: Reaching into service registry. Not only is it ugly, but it
        # only works for the default implementation. Need to make this kind
        # of information available via the public API.
        all_services = list(self.service_registry._services.items())

        protocols = {}
        for service_id, (protocol_name, obj, properties) in all_services:
            protocol = protocols.get(protocol_name)
            if protocol is None:
                protocol = ProtocolModel(name=protocol_name)
                protocols[protocol_name] = protocol

            service_model = ServiceModel(
                id         = service_id,
                obj        = obj,
                properties = properties
            )

            protocol.services.append(service_model)

        return list(protocols.values())


class ProtocolModelTreeNode(TreeNode):
    """ A tree node for a protocol model. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return True

    def get_children(self, obj):
        """ Get the object's children. """

        svtno = SingleValueTreeNodeObject()

        nodes = []
        for service in obj.services:
            node = svtno.node_for(repr(service.obj), service.obj)
            node._protocol_   = obj.name
            node._service_id_ = service.id
            nodes.append(node)

        return nodes


service_registry_browser_tree_nodes = [
    TreeNode(
        node_for  = [ServiceRegistryModel],
        auto_open = True,
        label     = '=Services',
        children  = 'protocols',
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),

    ProtocolModelTreeNode(
        node_for  = [ProtocolModel],
        auto_open = True,
        label     = 'name',
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),
] + value_tree_nodes

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The address of a commuications endpoint. """


# Enthought library imports.
from traits.api import Any, HasTraits, Str


class Address(HasTraits):
    """ The address of a communications end-point.

    It contains a type that describes the communication mechanism, and the
    actual address content.

    """

    # The type of the address.
    type = Str

    # The actual content.
    content = Any

#### EOF ######################################################################



# Enthought library imports.
from traits.api import Interface, Str, List, Instance, provides, HasTraits

from omnivore.framework.document import BaseDocument


class IFileRecognizer(Interface):
    """File recognizers must implement this AND be added to the list of known
    recognizers via a plugin that contributes to 'omnivore.file_recognizer'.
    
    """
    # The ID used in ordering recognizers.  This should correspond to a MIME
    # type that can be detected with this recognizer (if more than one MIME
    # type can be recognized, the most common MIME type can be used or a MIME-
    # like string containing the top-level media type and a descriptive subtype
    # may be used.  E.g.  "image/common" is used by the ImageRecognizer
    # plugin).
    id = Str

    # The recognizer will be processed after the item with this ID.
    after = Str

    # The recognizer will be processed before the item with this ID.
    before = Str

    def identify(self, guess):
        """Return a MIME type if the FileGuess can be identified.
        
        If the file type is not matched by this recognizer, returns None
        """


class RecognizerBase(HasTraits):
    def load(self, guess):
        doc = BaseDocument(metadata=guess.metadata, bytes=guess.numpy)
        doc.load_metadata(guess)
        return doc


class IFileRecognizerDriver(Interface):
    """ File type identifier service.
    
    """
    recognizers = List(Instance(IFileRecognizer))

    def recognize(self, guess):
        """Attempt to set the mime type of a FileGuess.
        """



#!/usr/bin/env python

"""
The basic test of the UGrid object.

More specific functionality is other test modules.

"""

from __future__ import (absolute_import, division, print_function)

from pyugrid import UGrid
from pyugrid.ugrid import IND_DT, NODE_DT

# FIXME: Break `test_full_set` into small unittests and check if the grid here
# is the same as `two_triangles`. If so use that.
# Some sample grid data: about the simplest triangle grid possible.
# 4 nodes, two triangles, five edges.

nodes = [(0.1, 0.1),
         (2.1, 0.1),
         (1.1, 2.1),
         (3.1, 2.1)]

faces = [(0, 1, 2),
         (1, 3, 2)]

edges = [(0, 1),
         (1, 3),
         (3, 2),
         (2, 0),
         (1, 2)]

boundaries = [(0, 1),
              (0, 2),
              (1, 3),
              (2, 3)]


def test_full_set():
    grid = UGrid(nodes=nodes,
                 faces=faces,
                 edges=edges,
                 boundaries=boundaries,
                 )

    # Check the dtype of key objects.
    # Implicitly makes sure they are numpy arrays (or array-like).
    assert grid.num_vertices == 3

    assert grid.nodes.dtype == NODE_DT
    assert grid.faces.dtype == IND_DT
    assert grid.edges.dtype == IND_DT
    assert grid.boundaries.dtype == IND_DT

    # Check shape of grid arrays.
    assert len(grid.nodes.shape) == 2
    assert len(grid.faces.shape) == 2
    assert len(grid.edges.shape) == 2
    assert len(grid.boundaries.shape) == 2

    assert grid.nodes.shape[1] == 2
    assert grid.faces.shape[1] == 3
    assert grid.edges.shape[1] == 2
    assert grid.boundaries.shape[1] == 2



"""Sample panes for Image Editor

"""
# Enthought library imports.
from pyface.tasks.api import DockPane


class Pane1(DockPane):
    #### TaskPane interface ###################################################

    id = 'image_edit.pane1'
    name = 'Pane 1'


class Pane2(DockPane):
    #### TaskPane interface ###################################################

    id = 'image_edit.pane2'
    name = 'Pane 2'


class Pane3(DockPane):
    #### TaskPane interface ###################################################

    id = 'image_edit.pane3'
    name = 'Pane 3'



""" Tests for the 'Egg Basket' plugin manager. """
from __future__ import print_function

import glob
import sys
from os.path import basename, dirname, join
import pkg_resources
import shutil
import tempfile

from envisage.egg_basket_plugin_manager import EggBasketPluginManager
from traits.testing.unittest_tools import unittest


class EggBasketPluginManagerTestCase(unittest.TestCase):
    """ Tests for the 'Egg Basket' plugin manager. """

    #### 'unittest.TestCase' protocol #########################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        # The location of the 'eggs' test data directory.
        self.eggs_dir = join(dirname(__file__), 'eggs')
        self.bad_eggs_dir = join(dirname(__file__), 'bad_eggs')

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """
        # Undo any side-effects: egg_basket_plugin_manager modifies sys.path.
        sys_path = []
        for path in sys.path:
            if self.bad_eggs_dir not in path:
                sys_path.append(path)
            else:
                print("Removed", path)
        sys.path = sys_path

        # `envisage.egg_utils.get_entry_points_in_egg_order` modifies the
        # global working set.
        pkg_resources.working_set = pkg_resources.WorkingSet()

        return

    #### Tests ################################################################

    def test_find_plugins_in_eggs_on_the_plugin_path(self):

        plugin_manager = EggBasketPluginManager(
            plugin_path=[self.eggs_dir]
        )

        ids = [plugin.id for plugin in plugin_manager]
        self.assertEqual(len(ids), 3)
        self.assertIn('acme.foo', ids)
        self.assertIn('acme.bar', ids)
        self.assertIn('acme.baz', ids)

        return

    def test_only_find_plugins_whose_ids_are_in_the_include_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        include = ['acme.foo', 'acme.bar']

        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.eggs_dir],
            include     = include
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.foo', 'acme.bar']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_only_find_plugins_matching_a_wildcard_in_the_include_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        include = ['acme.b*']

        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.eggs_dir],
            include     = include
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.bar', 'acme.baz']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_ignore_plugins_whose_ids_are_in_the_exclude_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        exclude = ['acme.foo', 'acme.baz']

        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.eggs_dir],
            exclude     = exclude
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.bar']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_ignore_plugins_matching_a_wildcard_in_the_exclude_list(self):

        # Note that the items in the list use the 'fnmatch' syntax for matching
        # plugins Ids.
        exclude = ['acme.b*']

        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.eggs_dir],
            exclude     = exclude
        )

        # The Ids of the plugins that we expect the plugin manager to find.
        expected = ['acme.foo']

        # Make sure the plugin manager found only the required plugins and that
        # it starts and stops them correctly..
        self._test_start_and_stop(plugin_manager, expected)

        return

    def test_reflect_changes_to_the_plugin_path(self):
        plugin_manager = EggBasketPluginManager()
        ids = [plugin.id for plugin in plugin_manager]
        self.assertEqual(len(ids), 0)

        plugin_manager.plugin_path.append(self.eggs_dir)
        ids = [plugin.id for plugin in plugin_manager]
        self.assertEqual(len(ids), 3)
        self.assertIn('acme.foo', ids)
        self.assertIn('acme.bar', ids)
        self.assertIn('acme.baz', ids)

        del plugin_manager.plugin_path[0]
        ids = [plugin.id for plugin in plugin_manager]
        self.assertEqual(len(ids), 0)

        return

    def test_ignore_broken_plugins_raises_exceptions_by_default(self):
        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.bad_eggs_dir, self.eggs_dir],
        )
        self.assertRaises(ImportError, list, plugin_manager)

        return

    def test_ignore_broken_plugins_loads_good_plugins(self):
        data = {'count': 0}

        def on_broken_plugin(ep, exc):
            data['count'] += 1
            data['entry_point'] = ep
            data['exc'] = exc

        plugin_manager = EggBasketPluginManager(
            plugin_path           = [self.bad_eggs_dir, self.eggs_dir],
            on_broken_plugin      = on_broken_plugin,
        )

        ids = [plugin.id for plugin in plugin_manager]
        self.assertEqual(len(ids), 3)
        self.assertIn('acme.foo', ids)
        self.assertIn('acme.bar', ids)
        self.assertIn('acme.baz', ids)

        self.assertEqual(data['count'], 1)
        self.assertEqual(data['entry_point'].name, 'acme.bad')
        exc = data['exc']
        self.assertTrue(isinstance(exc, ImportError))

        return

    def test_ignore_broken_distributions_raises_exceptions_by_default(self):
        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.bad_eggs_dir,
                self._create_broken_distribution_eggdir('acme.foo*.egg')],
        )
        self.assertRaises(SystemError, iter, plugin_manager)

        return

    def test_ignore_broken_distributions_loads_good_distributions(self):
        data = {'count':0}
        def on_broken_distribution(dist, exc):
            data['count'] += 1
            data['distribution'] = dist
            data['exc'] = exc

        plugin_manager = EggBasketPluginManager(
            plugin_path = [self.eggs_dir,
                self._create_broken_distribution_eggdir('acme.foo*.egg')],
            on_broken_distribution = on_broken_distribution,
        )

        ids = [plugin.id for plugin in plugin_manager]
        self.assertEqual(len(ids), 3)
        self.assertIn('acme.foo', ids)
        self.assertIn('acme.bar', ids)
        self.assertIn('acme.baz', ids)

        self.assertEqual(data['count'], 1)
        self.assertEqual(data['distribution'].project_name, 'acme.foo')
        exc = data['exc']
        self.assertTrue(isinstance(exc, pkg_resources.VersionConflict))

        return

    #### Private protocol #####################################################

    def _test_start_and_stop(self, plugin_manager, expected):
        """ Make sure the plugin manager starts and stops the expected plugins.

        """

        # Make sure the plugin manager found only the required plugins.
        self.assertEqual(expected, [plugin.id for plugin in plugin_manager])

        # Start the plugin manager. This starts all of the plugin manager's
        # plugins.
        plugin_manager.start()

        # Make sure all of the the plugins were started.
        for id in expected:
            plugin = plugin_manager.get_plugin(id)
            self.assertNotEqual(None, plugin)
            self.assertEqual(True, plugin.started)

        # Stop the plugin manager. This stops all of the plugin manager's
        # plugins.
        plugin_manager.stop()

        # Make sure all of the the plugins were stopped.
        for id in expected:
            plugin = plugin_manager.get_plugin(id)
            self.assertNotEqual(None, plugin)
            self.assertEqual(True, plugin.stopped)

        return

    def _create_broken_distribution_eggdir(self, egg_pat, replacement=None):
        """ Copy a good egg to a different version egg name in a new temp dir
        and return the new directory.

        Parameters
        ----------
        egg_pat: a glob pattern for the egg in `self.egg_dir` eg 'foo.bar*.egg'
        replacement: a string replacement for the version part of egg name.
            If None, '1' is appended to the original version.

        Returns
        -------
        The newly created dir where the new broken egg is copied.
        Adding this dir to plugin_path will cause VersionConflict
        on trying to load distributions.
        """
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)

        # Copy the egg to the temp dir and rename it
        eggs = glob.glob(join(self.eggs_dir, egg_pat))
        for egg in eggs:
            egg_name = basename(egg)
            split_name = egg_name.split('-')
            if replacement is None:
                split_name[1] += '1'
            else:
                split_name[1] = replacement
            new_name = '-'.join(split_name)
            shutil.copy(egg, join(tmpdir, new_name))

        return tmpdir


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



"""
fs.wrapfs.hidedotfilesfs
========================

An FS wrapper class for hiding dot-files in directory listings.

"""

from fs.wrapfs import WrapFS
from fs.path import *
from fnmatch import fnmatch


class HideDotFilesFS(WrapFS):
    """FS wrapper class that hides dot-files in directory listings.

    The listdir() function takes an extra keyword argument 'hidden'
    indicating whether hidden dot-files should be included in the output.
    It is False by default.
    """

    def is_hidden(self, path):
        """Check whether the given path should be hidden."""
        return path and basename(path)[0] == "."

    def _encode(self, path):
        return path

    def _decode(self, path):
        return path

    def listdir(self, path="", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False, hidden=False):
        kwds = dict(wildcard=wildcard,
                    full=full,
                    absolute=absolute,
                    dirs_only=dirs_only,
                    files_only=files_only)
        entries = self.wrapped_fs.listdir(path,**kwds)
        if not hidden:
            entries = [e for e in entries if not self.is_hidden(e)]
        return entries

    def ilistdir(self, path="", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False, hidden=False):
        kwds = dict(wildcard=wildcard,
                    full=full,
                    absolute=absolute,
                    dirs_only=dirs_only,
                    files_only=files_only)
        for e in self.wrapped_fs.ilistdir(path,**kwds):
            if hidden or not self.is_hidden(e):
                yield e

    def walk(self, path="/", wildcard=None, dir_wildcard=None, search="breadth",hidden=False):
        if search == "breadth":
            dirs = [path]
            while dirs:
                current_path = dirs.pop()
                paths = []
                for filename in self.listdir(current_path,hidden=hidden):
                    path = pathjoin(current_path, filename)
                    if self.isdir(path):
                        if dir_wildcard is not None:
                            if fnmatch(path, dir_wildcard):
                                dirs.append(path)
                        else:
                            dirs.append(path)
                    else:
                        if wildcard is not None:
                            if fnmatch(path, wildcard):
                                paths.append(filename)
                        else:
                            paths.append(filename)
                yield (current_path, paths)
        elif search == "depth":
            def recurse(recurse_path):
                for path in self.listdir(recurse_path, wildcard=dir_wildcard, full=True, dirs_only=True,hidden=hidden):
                    for p in recurse(path):
                        yield p
                yield (recurse_path, self.listdir(recurse_path, wildcard=wildcard, files_only=True,hidden=hidden))
            for p in recurse(path):
                yield p
        else:
            raise ValueError("Search should be 'breadth' or 'depth'")


    def isdirempty(self, path):
        path = normpath(path)
        iter_dir = iter(self.listdir(path,hidden=True))
        try:
            iter_dir.next()
        except StopIteration:
            return True
        return False





# -*- coding: UTF-8 -*-
# =============================================================================
# Copyright (C) 2012 Brad Hards <bradh@frogmouth.net>
#
# Based on wms.py, which has the following copyright statement:
# Copyright (c) 2004, 2006 Sean C. Gillies
# Copyright (c) 2005 Nuxeo SARL <http://nuxeo.com>
#
# Authors : Sean Gillies <sgillies@frii.com>
#           Julien Anguenot <ja@nuxeo.com>
#
# Contact email: sgillies@frii.com
# =============================================================================

"""

Abstract
--------
The wmts module of the OWSlib package provides client-side functionality
for fetching tiles from an OGC Web Map Tile Service (WMTS)


Disclaimer
----------
PLEASE NOTE: the owslib wmts module should be considered in early-beta
state: it has been tested against only one WMTS server (NASA EODSIS).
More extensive testing is needed and feedback (to bradh@frogmouth.net)
would be appreciated.

"""

from __future__ import (absolute_import, division, print_function)

import warnings
import six
from six.moves import filter
try:                    # Python 3
    from urllib.parse import (urlencode, urlparse, urlunparse, parse_qs,
                              ParseResult)
except ImportError:      # Python 2
    from urllib import urlencode
    from urlparse import urlparse, urlunparse, parse_qs, ParseResult
from .etree import etree
from .util import openURL, testXMLValue, getXMLInteger
from .fgdc import Metadata
from .iso import MD_Metadata
from .ows import ServiceProvider, ServiceIdentification, OperationsMetadata


_OWS_NS = '{http://www.opengis.net/ows/1.1}'
_WMTS_NS = '{http://www.opengis.net/wmts/1.0}'
_XLINK_NS = '{http://www.w3.org/1999/xlink}'

_ABSTRACT_TAG = _OWS_NS + 'Abstract'
_IDENTIFIER_TAG = _OWS_NS + 'Identifier'
_LOWER_CORNER_TAG = _OWS_NS + 'LowerCorner'
_OPERATIONS_METADATA_TAG = _OWS_NS + 'OperationsMetadata'
_SERVICE_IDENTIFICATION_TAG = _OWS_NS + 'ServiceIdentification'
_SERVICE_PROVIDER_TAG = _OWS_NS + 'ServiceProvider'
_SUPPORTED_CRS_TAG = _OWS_NS + 'SupportedCRS'
_TITLE_TAG = _OWS_NS + 'Title'
_UPPER_CORNER_TAG = _OWS_NS + 'UpperCorner'
_WGS84_BOUNDING_BOX_TAG = _OWS_NS + 'WGS84BoundingBox'

_CONTENTS_TAG = _WMTS_NS + 'Contents'
_FORMAT_TAG = _WMTS_NS + 'Format'
_INFO_FORMAT_TAG = _WMTS_NS + 'InfoFormat'
_LAYER_TAG = _WMTS_NS + 'Layer'
_LAYER_REF_TAG = _WMTS_NS + 'LayerRef'
_MATRIX_HEIGHT_TAG = _WMTS_NS + 'MatrixHeight'
_MATRIX_WIDTH_TAG = _WMTS_NS + 'MatrixWidth'
_MAX_TILE_COL_TAG = _WMTS_NS + 'MaxTileCol'
_MAX_TILE_ROW_TAG = _WMTS_NS + 'MaxTileRow'
_MIN_TILE_COL_TAG = _WMTS_NS + 'MinTileCol'
_MIN_TILE_ROW_TAG = _WMTS_NS + 'MinTileRow'
_RESOURCE_URL_TAG = _WMTS_NS + 'ResourceURL'
_SCALE_DENOMINATOR_TAG = _WMTS_NS + 'ScaleDenominator'
_SERVICE_METADATA_URL_TAG = _WMTS_NS + 'ServiceMetadataURL'
_STYLE_TAG = _WMTS_NS + 'Style'
_THEME_TAG = _WMTS_NS + 'Theme'
_THEMES_TAG = _WMTS_NS + 'Themes'
_TILE_HEIGHT_TAG = _WMTS_NS + 'TileHeight'
_TILE_MATRIX_SET_LINK_TAG = _WMTS_NS + 'TileMatrixSetLink'
_TILE_MATRIX_SET_TAG = _WMTS_NS + 'TileMatrixSet'
_TILE_MATRIX_SET_LIMITS_TAG = _WMTS_NS + 'TileMatrixSetLimits'
_TILE_MATRIX_LIMITS_TAG = _WMTS_NS + 'TileMatrixLimits'
_TILE_MATRIX_TAG = _WMTS_NS + 'TileMatrix'
_TILE_WIDTH_TAG = _WMTS_NS + 'TileWidth'
_TOP_LEFT_CORNER_TAG = _WMTS_NS + 'TopLeftCorner'

_HREF_TAG = _XLINK_NS + 'href'


class ServiceException(Exception):
    """WMTS ServiceException

    Attributes:
        message -- short error message
        xml  -- full xml error message from server
    """

    def __init__(self, message, xml):
        self.message = message
        self.xml = xml

    def __str__(self):
        return repr(self.message)


class CapabilitiesError(Exception):
    pass


class WebMapTileService(object):
    """Abstraction for OGC Web Map Tile Service (WMTS).

    Implements IWebMapService.
    """

    def __getitem__(self, name):
        '''Check contents dictionary to allow dict like access to
        service layers'''
        if name in self.__getattribute__('contents'):
            return self.__getattribute__('contents')[name]
        else:
            raise KeyError("No content named %s" % name)

    def __init__(self, url, version='1.0.0', xml=None, username=None,
                 password=None, parse_remote_metadata=False,
                 vendor_kwargs=None):
        """Initialize.

        Parameters
        ----------
        url : string
            Base URL for the WMTS service.
        version : string
            Optional WMTS version. Defaults to '1.0.0'.
        xml : string
            Optional XML content to use as the content for the initial
            GetCapabilities request. Typically only used for testing.
        username : string
            Optional user name for authentication.
        password : string
            Optional password for authentication.
        parse_remote_metadata: string
            Currently unused.
        vendor_kwargs : dict
            Optional vendor-specific parameters to be included in all
            requests.

        """
        self.url = url
        self.username = username
        self.password = password
        self.version = version
        self.vendor_kwargs = vendor_kwargs
        self._capabilities = None

        # Authentication handled by Reader
        reader = WMTSCapabilitiesReader(self.version, url=self.url,
                                        un=self.username, pw=self.password)

        if xml:  # read from stored xml
            self._capabilities = reader.readString(xml)
        else:  # read from server
            self._capabilities = reader.read(self.url, self.vendor_kwargs)

        # Avoid building capabilities metadata if the response is a
        # ServiceExceptionReport.
        # TODO: check if this needs a namespace
        se = self._capabilities.find('ServiceException')
        if se is not None:
            err_message = str(se.text).strip()
            raise ServiceException(err_message, xml)

        # build metadata objects
        self._buildMetadata(parse_remote_metadata)

    def _getcapproperty(self):
        if not self._capabilities:
            reader = WMTSCapabilitiesReader(
                self.version, url=self.url, un=self.username, pw=self.password
                )
            xml = reader.read(self.url, self.vendor_kwargs)
            self._capabilities = ServiceMetadata(xml)
        return self._capabilities

    def _buildMetadata(self, parse_remote_metadata=False):
        ''' set up capabilities metadata objects '''

        # serviceIdentification metadata
        serviceident = self._capabilities.find(_SERVICE_IDENTIFICATION_TAG)
        self.identification = ServiceIdentification(serviceident)

        # serviceProvider metadata
        serviceprov = self._capabilities.find(_SERVICE_PROVIDER_TAG)
        self.provider = ServiceProvider(serviceprov)

        # serviceOperations metadata
        self.operations = []
        for elem in self._capabilities.find(_OPERATIONS_METADATA_TAG)[:]:
            self.operations.append(OperationsMetadata(elem))

        # serviceContents metadata: our assumption is that services use
        # a top-level layer as a metadata organizer, nothing more.
        self.contents = {}
        caps = self._capabilities.find(_CONTENTS_TAG)

        def gather_layers(parent_elem, parent_metadata):
            for index, elem in enumerate(parent_elem.findall(_LAYER_TAG)):
                cm = ContentMetadata(
                    elem, parent=parent_metadata, index=index+1,
                    parse_remote_metadata=parse_remote_metadata)
                if cm.id:
                    if cm.id in self.contents:
                        raise KeyError('Content metadata for layer "%s" '
                                       'already exists' % cm.id)
                    self.contents[cm.id] = cm
                gather_layers(elem, cm)
        gather_layers(caps, None)

        self.tilematrixsets = {}
        for elem in caps.findall(_TILE_MATRIX_SET_TAG):
            tms = TileMatrixSet(elem)
            if tms.identifier:
                if tms.identifier in self.tilematrixsets:
                    raise KeyError('TileMatrixSet with identifier "%s" '
                                   'already exists' % tms.identifier)
                self.tilematrixsets[tms.identifier] = tms

        self.themes = {}
        for elem in self._capabilities.findall(_THEMES_TAG + '/' + _THEME_TAG):
            theme = Theme(elem)
            if theme.identifier:
                if theme.identifier in self.themes:
                    raise KeyError('Theme with identifier "%s" already exists'
                                   % theme.identifier)
                self.themes[theme.identifier] = theme

        serviceMetadataURL = self._capabilities.find(_SERVICE_METADATA_URL_TAG)
        if serviceMetadataURL is not None:
            self.serviceMetadataURL = serviceMetadataURL.attrib[_HREF_TAG]
        else:
            self.serviceMetadataURL = None

    def items(self):
        '''supports dict-like items() access'''
        items = []
        for item in self.contents:
            items.append((item, self.contents[item]))
        return items

    def buildTileRequest(self, layer=None, style=None, format=None,
                         tilematrixset=None, tilematrix=None, row=None,
                         column=None, **kwargs):
        """Return the URL-encoded parameters for a GetTile request.

        Parameters
        ----------
        layer : string
            Content layer name.
        style : string
            Optional style name. Defaults to the first style defined for
            the relevant layer in the GetCapabilities response.
        format : string
            Optional output image format,  such as 'image/jpeg'.
            Defaults to the first format defined for the relevant layer
            in the GetCapabilities response.
        tilematrixset : string
            Optional name of tile matrix set to use.
            Defaults to the first tile matrix set defined for the
            relevant layer in the GetCapabilities response.
        tilematrix : string
            Name of the tile matrix to use.
        row : integer
            Row index of tile to request.
        column : integer
            Column index of tile to request.
        **kwargs : extra arguments
            anything else e.g. vendor specific parameters

        Example
        -------
            >>> url = 'http://map1c.vis.earthdata.nasa.gov/wmts-geo/wmts.cgi'
            >>> wmts = WebMapTileService(url)
            >>> wmts.buildTileRequest(layer='VIIRS_CityLights_2012',
            ...                       tilematrixset='EPSG4326_500m',
            ...                       tilematrix='6',
            ...                       row=4, column=4)
            'SERVICE=WMTS&REQUEST=GetTile&VERSION=1.0.0&\
LAYER=VIIRS_CityLights_2012&STYLE=default&TILEMATRIXSET=EPSG4326_500m&\
TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'

        """
        request = {'version': self.version, 'request': 'GetTile'}

        if (layer is None):
            raise ValueError("layer is mandatory (cannot be None)")
        if style is None:
            style = list(self[layer].styles.keys())[0]
        if format is None:
            format = self[layer].formats[0]
        if tilematrixset is None:
            tilematrixset = sorted(self[layer].tilematrixsetlinks.keys())[0]
        if tilematrix is None:
            msg = 'tilematrix (zoom level) is mandatory (cannot be None)'
            raise ValueError(msg)
        if row is None:
                raise ValueError("row is mandatory (cannot be None)")
        if column is None:
                raise ValueError("column is mandatory (cannot be None)")

        request = list()
        request.append(('SERVICE', 'WMTS'))
        request.append(('REQUEST', 'GetTile'))
        request.append(('VERSION', '1.0.0'))
        request.append(('LAYER', layer))
        request.append(('STYLE', style))
        request.append(('TILEMATRIXSET', tilematrixset))
        request.append(('TILEMATRIX', tilematrix))
        request.append(('TILEROW', str(row)))
        request.append(('TILECOL', str(column)))
        request.append(('FORMAT', format))

        for key, value in six.iteritems(kwargs):
            request.append((key, value))

        data = urlencode(request, True)
        return data

    def gettile(self, base_url=None, layer=None, style=None, format=None,
                tilematrixset=None, tilematrix=None, row=None, column=None,
                **kwargs):
        """Return a tile from the WMTS.

        Returns the tile image as a file-like object.

        Parameters
        ----------
        base_url : string
            Optional URL for request submission. Defaults to the URL of
            the GetTile operation as declared in the GetCapabilities
            response.
        layer : string
            Content layer name.
        style : string
            Optional style name. Defaults to the first style defined for
            the relevant layer in the GetCapabilities response.
        format : string
            Optional output image format,  such as 'image/jpeg'.
            Defaults to the first format defined for the relevant layer
            in the GetCapabilities response.
        tilematrixset : string
            Optional name of tile matrix set to use.
            Defaults to the first tile matrix set defined for the
            relevant layer in the GetCapabilities response.
        tilematrix : string
            Name of the tile matrix to use.
        row : integer
            Row index of tile to request.
        column : integer
            Column index of tile to request.
        **kwargs : extra arguments
            anything else e.g. vendor specific parameters

        Example
        -------
            >>> url = 'http://map1c.vis.earthdata.nasa.gov/wmts-geo/wmts.cgi'
            >>> wmts = WebMapTileService(url)
            >>> img = wmts.gettile(layer='VIIRS_CityLights_2012',\
                                   tilematrixset='EPSG4326_500m',\
                                   tilematrix='6',\
                                   row=4, column=4)
            >>> out = open('tile.jpg', 'wb')
            >>> bytes_written = out.write(img.read())
            >>> out.close()

        """
        vendor_kwargs = self.vendor_kwargs or {}
        vendor_kwargs.update(kwargs)
        data = self.buildTileRequest(layer, style, format, tilematrixset,
                                     tilematrix, row, column, **vendor_kwargs)

        if base_url is None:
            base_url = self.url
            try:
                methods = self.getOperationByName('GetTile').methods
                get_verbs = [x for x in methods
                             if x.get('type').lower() == 'get']
                if len(get_verbs) > 1:
                    # Filter by constraints
                    base_url = next(
                        x for x in filter(
                            list,
                            ([pv.get('url')
                                for const in pv.get('constraints')
                                if 'kvp' in [x.lower() for x in const.values]]
                             for pv in get_verbs if pv.get('constraints'))))[0]
                elif len(get_verbs) == 1:
                    base_url = get_verbs[0].get('url')
            except StopIteration:
                pass
        u = openURL(base_url, data, username=self.username,
                    password=self.password)

        # check for service exceptions, and return
        if u.info()['Content-Type'] == 'application/vnd.ogc.se_xml':
            se_xml = u.read()
            se_tree = etree.fromstring(se_xml)
            err_message = six.text_type(se_tree.find('ServiceException').text)
            raise ServiceException(err_message.strip(), se_xml)
        return u

    def getServiceXML(self):
        xml = None
        if self._capabilities is not None:
            xml = etree.tostring(self._capabilities)
        return xml

    def getfeatureinfo(self):
        raise NotImplementedError

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)


class TileMatrixSet(object):
    '''Holds one TileMatrixSet'''
    def __init__(self, elem):
        if elem.tag != _TILE_MATRIX_SET_TAG:
            raise ValueError('%s should be a TileMatrixSet' % (elem,))
        self.identifier = testXMLValue(elem.find(_IDENTIFIER_TAG)).strip()
        self.crs = testXMLValue(elem.find(_SUPPORTED_CRS_TAG)).strip()
        if self.crs is None or self.identifier is None:
            raise ValueError('%s incomplete TileMatrixSet' % (elem,))
        self.tilematrix = {}
        for tilematrix in elem.findall(_TILE_MATRIX_TAG):
            tm = TileMatrix(tilematrix)
            if tm.identifier:
                if tm.identifier in self.tilematrix:
                    raise KeyError('TileMatrix with identifier "%s" '
                                   'already exists' % tm.identifier)
                self.tilematrix[tm.identifier] = tm


class TileMatrix(object):
    '''Holds one TileMatrix'''
    def __init__(self, elem):
        if elem.tag != _TILE_MATRIX_TAG:
            raise ValueError('%s should be a TileMatrix' % (elem,))
        self.identifier = testXMLValue(elem.find(_IDENTIFIER_TAG)).strip()
        sd = testXMLValue(elem.find(_SCALE_DENOMINATOR_TAG))
        if sd is None:
            raise ValueError('%s is missing ScaleDenominator' % (elem,))
        self.scaledenominator = float(sd)
        tl = testXMLValue(elem.find(_TOP_LEFT_CORNER_TAG))
        if tl is None:
            raise ValueError('%s is missing TopLeftCorner' % (elem,))
        (lon, lat) = tl.split(" ")
        self.topleftcorner = (float(lon), float(lat))
        width = testXMLValue(elem.find(_TILE_WIDTH_TAG))
        height = testXMLValue(elem.find(_TILE_HEIGHT_TAG))
        if (width is None) or (height is None):
            msg = '%s is missing TileWidth and/or TileHeight' % (elem,)
            raise ValueError(msg)
        self.tilewidth = int(width)
        self.tileheight = int(height)
        mw = testXMLValue(elem.find(_MATRIX_WIDTH_TAG))
        mh = testXMLValue(elem.find(_MATRIX_HEIGHT_TAG))
        if (mw is None) or (mh is None):
            msg = '%s is missing MatrixWidth and/or MatrixHeight' % (elem,)
            raise ValueError(msg)
        self.matrixwidth = int(mw)
        self.matrixheight = int(mh)


class Theme:
    """
    Abstraction for a WMTS theme
    """
    def __init__(self, elem):
        if elem.tag != _THEME_TAG:
            raise ValueError('%s should be a Theme' % (elem,))
        self.identifier = testXMLValue(elem.find(_IDENTIFIER_TAG)).strip()
        title = testXMLValue(elem.find(_TITLE_TAG))
        if title is not None:
            self.title = title.strip()
        else:
            self.title = None
        abstract = testXMLValue(elem.find(_ABSTRACT_TAG))
        if abstract is not None:
            self.abstract = abstract.strip()
        else:
            self.abstract = None

        self.layerRefs = []
        layerRefs = elem.findall(_LAYER_REF_TAG)
        for layerRef in layerRefs:
            if layerRef.text is not None:
                self.layerRefs.append(layerRef.text)


class TileMatrixLimits(object):
    """
    Represents a WMTS TileMatrixLimits element.

    """
    def __init__(self, elem):
        if elem.tag != _TILE_MATRIX_LIMITS_TAG:
            raise ValueError('%s should be a TileMatrixLimits' % elem)

        tm = elem.find(_TILE_MATRIX_TAG)
        if tm is None:
            raise ValueError('Missing TileMatrix in %s' % elem)
        self.tilematrix = tm.text.strip()

        self.mintilerow = getXMLInteger(elem, _MIN_TILE_ROW_TAG)
        self.maxtilerow = getXMLInteger(elem, _MAX_TILE_ROW_TAG)
        self.mintilecol = getXMLInteger(elem, _MIN_TILE_COL_TAG)
        self.maxtilecol = getXMLInteger(elem, _MAX_TILE_COL_TAG)

    def __repr__(self):
        fmt = ('<TileMatrixLimits: {self.tilematrix}'
               ', minRow={self.mintilerow}, maxRow={self.maxtilerow}'
               ', minCol={self.mintilecol}, maxCol={self.maxtilecol}>')
        return fmt.format(self=self)


class TileMatrixSetLink(object):
    """
    Represents a WMTS TileMatrixSetLink element.

    """
    @staticmethod
    def from_elements(link_elements):
        """
        Return a list of TileMatrixSetLink instances derived from the
        given list of <TileMatrixSetLink> XML elements.

        """
        # NB. The WMTS spec is contradictory re. the multiplicity
        # relationships between Layer and TileMatrixSetLink, and
        # TileMatrixSetLink and tileMatrixSet (URI).
        # Try to figure out which model has been used by the server.
        links = []
        for link_element in link_elements:
            matrix_set_elements = link_element.findall(_TILE_MATRIX_SET_TAG)
            if len(matrix_set_elements) == 0:
                raise ValueError('Missing TileMatrixSet in %s' % link_element)
            elif len(matrix_set_elements) > 1:
                set_limits_elements = link_element.findall(
                    _TILE_MATRIX_SET_LIMITS_TAG)
                if set_limits_elements:
                    raise ValueError('Multiple instances of TileMatrixSet'
                                     ' plus TileMatrixSetLimits in %s' %
                                     link_element)
                for matrix_set_element in matrix_set_elements:
                    uri = matrix_set_element.text.strip()
                    links.append(TileMatrixSetLink(uri))
            else:
                uri = matrix_set_elements[0].text.strip()

                tilematrixlimits = {}
                path = '%s/%s' % (_TILE_MATRIX_SET_LIMITS_TAG,
                                  _TILE_MATRIX_LIMITS_TAG)
                for limits_element in link_element.findall(path):
                    tml = TileMatrixLimits(limits_element)
                    if tml.tilematrix:
                        if tml.tilematrix in tilematrixlimits:
                            msg = ('TileMatrixLimits with tileMatrix "%s" '
                                   'already exists' % tml.tilematrix)
                            raise KeyError(msg)
                        tilematrixlimits[tml.tilematrix] = tml

                links.append(TileMatrixSetLink(uri, tilematrixlimits))
        return links

    def __init__(self, tilematrixset, tilematrixlimits=None):
        self.tilematrixset = tilematrixset

        if tilematrixlimits is None:
            self.tilematrixlimits = {}
        else:
            self.tilematrixlimits = tilematrixlimits

    def __repr__(self):
        fmt = ('<TileMatrixSetLink: {self.tilematrixset}'
               ', tilematrixlimits={{...}}>')
        return fmt.format(self=self)


class ContentMetadata:
    """
    Abstraction for WMTS layer metadata.

    Implements IContentMetadata.
    """
    def __init__(self, elem, parent=None, index=0,
                 parse_remote_metadata=False):
        if elem.tag != _LAYER_TAG:
            raise ValueError('%s should be a Layer' % (elem,))

        self.parent = parent
        if parent:
            self.index = "%s.%d" % (parent.index, index)
        else:
            self.index = str(index)

        self.id = self.name = testXMLValue(elem.find(_IDENTIFIER_TAG))
        # title is mandatory property
        self.title = None
        title = testXMLValue(elem.find(_TITLE_TAG))
        if title is not None:
            self.title = title.strip()

        self.abstract = testXMLValue(elem.find(_ABSTRACT_TAG))

        # bboxes
        b = elem.find(_WGS84_BOUNDING_BOX_TAG)
        self.boundingBox = None
        if b is not None:
            lc = b.find(_LOWER_CORNER_TAG)
            uc = b.find(_UPPER_CORNER_TAG)
            ll = [float(s) for s in lc.text.split()]
            ur = [float(s) for s in uc.text.split()]
            self.boundingBoxWGS84 = (ll[0], ll[1], ur[0], ur[1])
        # TODO: there is probably some more logic here, and it should
        # probably be shared code

        self._tilematrixsets = [f.text.strip() for f in
                                elem.findall(_TILE_MATRIX_SET_LINK_TAG + '/' +
                                             _TILE_MATRIX_SET_TAG)]

        link_elements = elem.findall(_TILE_MATRIX_SET_LINK_TAG)
        tile_matrix_set_links = TileMatrixSetLink.from_elements(link_elements)
        self.tilematrixsetlinks = {}
        for tmsl in tile_matrix_set_links:
            if tmsl.tilematrixset:
                if tmsl.tilematrixset in self.tilematrixsetlinks:
                    raise KeyError('TileMatrixSetLink with tilematrixset "%s"'
                                   ' already exists' %
                                   tmsl.tilematrixset)
                self.tilematrixsetlinks[tmsl.tilematrixset] = tmsl

        self.resourceURLs = []
        for resourceURL in elem.findall(_RESOURCE_URL_TAG):
            resource = {}
            for attrib in ['format', 'resourceType', 'template']:
                resource[attrib] = resourceURL.attrib[attrib]
            self.resourceURLs.append(resource)

        # Styles
        self.styles = {}
        for s in elem.findall(_STYLE_TAG):
            style = {}
            isdefaulttext = s.attrib.get('isDefault')
            style['isDefault'] = (isdefaulttext == "true")
            identifier = s.find(_IDENTIFIER_TAG)
            if identifier is None:
                raise ValueError('%s missing identifier' % (s,))
            title = s.find(_TITLE_TAG)
            if title is not None:
                style['title'] = title.text
            self.styles[identifier.text] = style

        self.formats = [f.text for f in elem.findall(_FORMAT_TAG)]

        self.infoformats = [f.text for f in elem.findall(_INFO_FORMAT_TAG)]

        self.layers = []
        for child in elem.findall(_LAYER_TAG):
            self.layers.append(ContentMetadata(child, self))

    @property
    def tilematrixsets(self):
        # NB. This attribute has been superseeded by the
        # `tilematrixsetlinks` attribute defined below, but is included
        # for now to provide continuity.
        warnings.warn("The 'tilematrixsets' attribute has been deprecated"
                      " and will be removed in a future version of OWSLib."
                      " Please use 'tilematrixsetlinks' instead.")
        return self._tilematrixsets

    def __str__(self):
        return 'Layer Name: %s Title: %s' % (self.name, self.title)


class WMTSCapabilitiesReader:
    """Read and parse capabilities document into a lxml.etree infoset
    """

    def __init__(self, version='1.0.0', url=None, un=None, pw=None):
        """Initialize"""
        self.version = version
        self._infoset = None
        self.url = url
        self.username = un
        self.password = pw

    def capabilities_url(self, service_url, vendor_kwargs=None):
        """Return a capabilities url
        """
        # Ensure the 'service', 'request', and 'version' parameters,
        # and any vendor-specific parameters are included in the URL.
        pieces = urlparse(service_url)
        args = parse_qs(pieces.query)
        if 'service' not in args:
            args['service'] = 'WMTS'
        if 'request' not in args:
            args['request'] = 'GetCapabilities'
        if 'version' not in args:
            args['version'] = self.version
        if vendor_kwargs:
            args.update(vendor_kwargs)
        query = urlencode(args, doseq=True)
        pieces = ParseResult(pieces.scheme, pieces.netloc,
                             pieces.path, pieces.params,
                             query, pieces.fragment)
        return urlunparse(pieces)

    def read(self, service_url, vendor_kwargs=None):
        """Get and parse a WMTS capabilities document, returning an
        elementtree instance

        service_url is the base url, to which is appended the service,
        version, and request parameters. Optional vendor-specific
        parameters can also be supplied as a dict.
        """
        getcaprequest = self.capabilities_url(service_url, vendor_kwargs)

        # now split it up again to use the generic openURL function...
        spliturl = getcaprequest.split('?')
        u = openURL(spliturl[0], spliturl[1], method='Get',
                    username=self.username, password=self.password)
        return etree.fromstring(u.read())

    def readString(self, st):
        """Parse a WMTS capabilities document, returning an elementtree instance

        string should be an XML capabilities document
        """
        if not isinstance(st, str) and not isinstance(st, bytes):
            msg = 'String must be of type string or bytes, not %s' % type(st)
            raise ValueError(msg)
        return etree.fromstring(st)




from nose.tools import assert_equals

from traits.api import HasTraits, Font

from traitsui.tests._tools import *


@skip_if_not_null
def test_font_trait_default():
    class Foo(HasTraits):
        font = Font()
    f = Foo()
    assert_equals(f.font, '10 pt Arial')


@skip_if_not_null
def test_font_trait_examples():
    """
    An assigned font string is parsed, and the substrings are
    put in the order: point size, family, style, weight, underline, facename

    The words 'pt, 'point' and 'family' are ignored.

    """
    class Foo(HasTraits):
        font = Font

    f = Foo(font='Qwerty 10')
    assert_equals(f.font, '10 pt Qwerty')

    f = Foo(font='nothing')
    assert_equals(f.font, 'nothing')

    f = Foo(font='swiss family arial')
    assert_equals(f.font, 'swiss arial')

    f = Foo(font='12 pt bold italic')
    assert_equals(f.font, '12 pt italic bold')

    f = Foo(font='123 Foo bar slant')
    assert_equals(f.font, '123 pt slant Foo bar')

    f = Foo(font='123 point Foo family bar slant')
    assert_equals(f.font, '123 pt slant Foo bar')

    f = Foo(font='16 xyzzy underline slant')
    assert_equals(f.font, '16 pt slant underline xyzzy')



""" Useful functions to do with types!

fixme: I don't really like collections of functions, but I'm not sure where
these should go? Class methods on 'TypeManager'?

"""


# Standard library imports.
from inspect import getclasstree


def sort_by_class_tree(objs):
    """ Sorts a list of objects by their class tree.

    Each item in the list should inherit from a common base class.

    The list is returned ordered from the most specific to the least specific.

    """

    # Since all objects must inherit from a common base class the list
    # returned by 'getclasstree' will be of length two:-
    #
    # The first element is a tuple of the form:-
    #
    # (CommonBaseClass, (HasTraits,))
    #
    # The second element is a possibly nested list containing all of the
    # classes derived from 'CommonBaseClass'.
    hierarchy = getclasstree([type(obj) for obj in objs])

    # Do an in-order traversal of the tree and return just the classes.
    #
    # This returns them ordered from least specific to most specific.
    classes = get_classes(hierarchy)

    # Note the reverse comparison (i.e., compare y with x). This is
    # because we want to return the classes ordered from the MOST
    # specfic to the least specific.
    objs.sort(key=lambda x: classes.index(type(x)), reverse=True)

    return


def get_classes(hierarchy):
    """ Walks a class hierarchy and returns all of the classes. """

    classes = []
    for item in hierarchy:
        if type(item) is tuple:
            classes.append(item[0])

        else:
            classes.extend(get_classes(item))

    return classes

#### EOF ######################################################################



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   08/29/2007

#-------------------------------------------------------------------------

""" Defines an ArrayViewEditor for displaying 1-d or 2-d arrays of values.
"""

#-- Imports --------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Instance, Property, List, Str, Bool, Font

from ..api import View, Item, TabularEditor, BasicEditorFactory

from ..tabular_adapter import TabularAdapter

from ..toolkit import toolkit_object

from ..ui_editor import UIEditor

#-- Tabular Adapter Definition -------------------------------------------


class ArrayViewAdapter(TabularAdapter):

    # Is the array 1D or 2D?
    is_2d = Bool(True)

    # Should array rows and columns be transposed:
    transpose = Bool(False)

    alignment = 'right'
    index_text = Property

    def _get_index_text(self):
        return str(self.row)

    def _get_content(self):
        if self.is_2d:
            return self.item[self.column_id]

        return self.item

    def get_item(self, object, trait, row):
        """ Returns the value of the *object.trait[row]* item.
        """
        if self.is_2d:
            if self.transpose:
                return getattr(object, trait)[:, row]

            return super(ArrayViewAdapter, self).get_item(object, trait,
                                                          row)

        return getattr(object, trait)[row]

    def len(self, object, trait):
        """ Returns the number of items in the specified *object.trait* list.
        """
        if self.transpose:
            return getattr(object, trait).shape[1]

        return super(ArrayViewAdapter, self).len(object, trait)

# Define the actual abstract Traits UI array view editor (each backend should
# implement its own editor that inherits from this class.


class _ArrayViewEditor(UIEditor):

    # Indicate that the editor is scrollable/resizable:
    scrollable = True

    # Should column titles be displayed:
    show_titles = Bool(False)

    # The tabular adapter being used for the editor view:
    adapter = Instance(ArrayViewAdapter)

    #-- Private Methods ------------------------------------------------------

    def _array_view(self):
        """ Return the view used by the editor.
        """
        return View(
            Item('object.object.' + self.name,
                 id='tabular_editor',
                 show_label=False,
                 editor=TabularEditor(show_titles=self.show_titles,
                                      editable=False,
                                      adapter=self.adapter)
                 ),
            id='array_view_editor',
            resizable=True
        )

    def init_ui(self, parent):
        """ Creates the Traits UI for displaying the array.
        """
        # Make sure that the value is an array of the correct shape:
        shape = self.value.shape
        len_shape = len(shape)
        if (len_shape == 0) or (len_shape > 2):
            raise ValueError("ArrayViewEditor can only display 1D or 2D "
                             "arrays")

        factory = self.factory
        cols = 1
        titles = factory.titles
        n = len(titles)
        self.show_titles = (n > 0)
        is_2d = (len_shape == 2)
        if is_2d:
            index = 1
            if factory.transpose:
                index = 0
            cols = shape[index]
            if self.show_titles:
                if n > cols:
                    titles = titles[:cols]
                elif n < cols:
                    if (cols % n) == 0:
                        titles, old_titles, i = [], titles, 0
                        while len(titles) < cols:
                            titles.extend('%s%d' % (title, i)
                                          for title in old_titles)
                            i += 1
                    else:
                        titles.extend([''] * (cols - n))
            else:
                titles = ['Data %d' % i for i in range(cols)]

        columns = [(title, i) for i, title in enumerate(titles)]

        if factory.show_index:
            columns.insert(0, ('Index', 'index'))

        self.adapter = ArrayViewAdapter(is_2d=is_2d,
                                        columns=columns,
                                        transpose=factory.transpose,
                                        format=factory.format,
                                        font=factory.font)

        return self.edit_traits(view='_array_view',
                                parent=parent,
                                kind='subpanel')

# Define the ArrayViewEditor class used by client code:


class ArrayViewEditor(BasicEditorFactory):

    # The editor implementation class:
    klass = Property

    # Should an index column be displayed:
    show_index = Bool(True)

    # List of (optional) column titles:
    titles = List(Str)

    # Should the array be logically transposed:
    transpose = Bool(False)

    # The format used to display each array element:
    format = Str('%s')

    # The font to use for displaying each array element:
    font = Font('Courier 10')

    def _get_klass(self):
        """ The class used to construct editor objects.
        """
        return toolkit_object('array_view_editor:_ArrayViewEditor')



""" The provider extension registry interface. """



# Local imports.
from .i_extension_registry import IExtensionRegistry


class IProviderExtensionRegistry(IExtensionRegistry):
    """ The provider extension registry interface. """

    def add_provider(self, provider):
        """ Add an extension provider.

        """

    def get_providers(self):
        """ Return all of the providers in the registry.

        """

    def remove_provider(self, provider):
        """ Remove an extension provider.

        Raise a 'ValueError' if the provider is not in the registry.

        """

#### EOF ######################################################################



""" Classes used to represent packages. """


# Standard library imports.
from os.path import join

# Enthought library imports.
from apptools.io.api import File
from traits.api import Instance, List, Str

# Local imports.
from .namespace import Namespace


class Package(Namespace):
    """ A package. """

    #### 'Package' interface ##################################################

    # The namespace that the package is defined in.
    #
    # fixme: This is always None for packages so why should we have this trait!
    # If I remember correctly it is something to do with trait identification!
    namespace = Instance(Namespace)

    # The package contents (modules and sub-packages).
    contents = List

    # The absolute filename of the package.
    filename = Str

    # The package name.
    name = Str

    # The package's parent package (None if this is a top level package).
    parent = Instance('Package')

    ###########################################################################
    # 'Package' interface.
    ###########################################################################

    def create_sub_package(self, name):
        """ Creates a sub-package with the specified name. """

        sub_package = File(join(self.filename, name))
        sub_package.create_package()

        return

    def delete_sub_package(self, name):
        """ Deletes the sub-package with the specified name. """

        sub_package = File(join(self.filename, name))
        sub_package.delete()

        return

    ###########################################################################
    # Private interface
    ###########################################################################

    def _name_changed(self, name):
        """ Called when the package name has been changed. """

        ##print '**** Package name changed', self.name

        return

#### EOF ######################################################################



"""
fs.expose.sftp
==============

Expose an FS object over SFTP (via paramiko).

This module provides the necessary interfaces to expose an FS object over
SFTP, plugging into the infrastructure provided by the 'paramiko' module.

For simple usage, the class 'BaseSFTPServer' provides an all-in-one server
class based on the standard SocketServer module.  Use it like so::

    server = BaseSFTPServer((hostname,port),fs)
    server.serve_forever()

Note that the base class allows UNAUTHENTICATED ACCESS by default.  For more
serious work you will probably want to subclass it and override methods such
as check_auth_password() and get_allowed_auths().

To integrate this module into an existing server framework based on paramiko,
the 'SFTPServerInterface' class provides a concrete implementation of the
paramiko.SFTPServerInterface protocol.  If you don't understand what this
is, you probably don't want to use it.

"""

from __future__ import with_statement

import os
import stat as statinfo
import time
import SocketServer
import threading

import paramiko

from fs.base import flags_to_mode
from fs.path import *
from fs.errors import *
from fs.local_functools import wraps
from fs.filelike import StringIO
from fs.utils import isdir


# Default host key used by BaseSFTPServer
#
DEFAULT_HOST_KEY = paramiko.RSAKey.from_private_key(StringIO(
    "-----BEGIN RSA PRIVATE KEY-----\n" \
    "MIICXgIBAAKCAIEAl7sAF0x2O/HwLhG68b1uG8KHSOTqe3Cdlj5i/1RhO7E2BJ4B\n" \
    "3jhKYDYtupRnMFbpu7fb21A24w3Y3W5gXzywBxR6dP2HgiSDVecoDg2uSYPjnlDk\n" \
    "HrRuviSBG3XpJ/awn1DObxRIvJP4/sCqcMY8Ro/3qfmid5WmMpdCZ3EBeC0CAwEA\n" \
    "AQKCAIBSGefUs5UOnr190C49/GiGMN6PPP78SFWdJKjgzEHI0P0PxofwPLlSEj7w\n" \
    "RLkJWR4kazpWE7N/bNC6EK2pGueMN9Ag2GxdIRC5r1y8pdYbAkuFFwq9Tqa6j5B0\n" \
    "GkkwEhrcFNBGx8UfzHESXe/uE16F+e8l6xBMcXLMJVo9Xjui6QJBAL9MsJEx93iO\n" \
    "zwjoRpSNzWyZFhiHbcGJ0NahWzc3wASRU6L9M3JZ1VkabRuWwKNuEzEHNK8cLbRl\n" \
    "TyH0mceWXcsCQQDLDEuWcOeoDteEpNhVJFkXJJfwZ4Rlxu42MDsQQ/paJCjt2ONU\n" \
    "WBn/P6iYDTvxrt/8+CtLfYc+QQkrTnKn3cLnAkEAk3ixXR0h46Rj4j/9uSOfyyow\n" \
    "qHQunlZ50hvNz8GAm4TU7v82m96449nFZtFObC69SLx/VsboTPsUh96idgRrBQJA\n" \
    "QBfGeFt1VGAy+YTLYLzTfnGnoFQcv7+2i9ZXnn/Gs9N8M+/lekdBFYgzoKN0y4pG\n" \
    "2+Q+Tlr2aNlAmrHtkT13+wJAJVgZATPI5X3UO0Wdf24f/w9+OY+QxKGl86tTQXzE\n" \
    "4bwvYtUGufMIHiNeWP66i6fYCucXCMYtx6Xgu2hpdZZpFw==\n" \
    "-----END RSA PRIVATE KEY-----\n"
))


def report_sftp_errors(func):
    """Decorator to catch and report FS errors as SFTP error codes.

    Any FSError exceptions are caught and translated into an appropriate
    return code, while other exceptions are passed through untouched.
    """
    @wraps(func)
    def wrapper(*args,**kwds):
        try:
            return func(*args, **kwds)
        except ResourceNotFoundError, e:
            return paramiko.SFTP_NO_SUCH_FILE
        except UnsupportedError, e:
            return paramiko.SFTP_OP_UNSUPPORTED
        except FSError, e:
            return paramiko.SFTP_FAILURE
    return wrapper


class SFTPServerInterface(paramiko.SFTPServerInterface):
    """SFTPServerInterface implementation that exposes an FS object.

    This SFTPServerInterface subclass expects a single additional argument,
    the fs object to be exposed.  Use it to set up a transport subsystem
    handler like so::

      t.set_subsystem_handler("sftp",SFTPServer,SFTPServerInterface,fs)

    If this all looks too complicated, you might consider the BaseSFTPServer
    class also provided by this module - it automatically creates the enclosing
    paramiko server infrastructure.
    """

    def __init__(self, server, fs, encoding=None, *args, **kwds):
        self.fs = fs
        if encoding is None:
            encoding = "utf8"
        self.encoding = encoding
        super(SFTPServerInterface,self).__init__(server, *args, **kwds)

    def close(self):
        # Close the pyfs file system and dereference it.
        self.fs.close()
        self.fs = None

    @report_sftp_errors
    def open(self, path, flags, attr):
        return SFTPHandle(self, path, flags)

    @report_sftp_errors
    def list_folder(self, path):
        if not isinstance(path, unicode):
            path = path.decode(self.encoding)
        stats = []
        for entry in self.fs.listdir(path, absolute=True):
            stat = self.stat(entry)
            if not isinstance(stat, int):
                stats.append(stat)
        return stats

    @report_sftp_errors
    def stat(self, path):
        if not isinstance(path, unicode):
            path = path.decode(self.encoding)

        info = self.fs.getinfo(path)

        stat = paramiko.SFTPAttributes()
        stat.filename = basename(path).encode(self.encoding)
        stat.st_size = info.get("size")

        if 'st_atime' in info:
            stat.st_atime = info.get('st_atime')
        elif 'accessed_time' in info:
            stat.st_atime = time.mktime(info.get("accessed_time").timetuple())

        if 'st_mtime' in info:
            stat.st_mtime = info.get('st_mtime')
        else:
            if 'modified_time' in info:
                stat.st_mtime = time.mktime(info.get("modified_time").timetuple())

        if isdir(self.fs, path, info):
            stat.st_mode = 0777 | statinfo.S_IFDIR
        else:
            stat.st_mode = 0777 | statinfo.S_IFREG
        return stat

    def lstat(self, path):
        return self.stat(path)

    @report_sftp_errors
    def remove(self, path):
        if not isinstance(path, unicode):
            path = path.decode(self.encoding)
        self.fs.remove(path)
        return paramiko.SFTP_OK

    @report_sftp_errors
    def rename(self, oldpath, newpath):
        if not isinstance(oldpath, unicode):
            oldpath = oldpath.decode(self.encoding)
        if not isinstance(newpath, unicode):
            newpath = newpath.decode(self.encoding)
        if self.fs.isfile(oldpath):
            self.fs.move(oldpath, newpath)
        else:
            self.fs.movedir(oldpath, newpath)
        return paramiko.SFTP_OK

    @report_sftp_errors
    def mkdir(self, path, attr):
        if not isinstance(path, unicode):
            path = path.decode(self.encoding)
        self.fs.makedir(path)
        return paramiko.SFTP_OK

    @report_sftp_errors
    def rmdir(self, path):
        if not isinstance(path, unicode):
            path = path.decode(self.encoding)
        self.fs.removedir(path)
        return paramiko.SFTP_OK

    def canonicalize(self, path):
        try:
            return abspath(normpath(path)).encode(self.encoding)
        except BackReferenceError:
            # If the client tries to use backrefs to escape root, gently
            # nudge them back to /.
            return '/'

    @report_sftp_errors
    def chattr(self, path, attr):
        #  f.truncate() is implemented by setting the size attr.
        #  Any other attr requests fail out.
        if attr._flags:
            if attr._flags != attr.FLAG_SIZE:
                raise UnsupportedError
            with self.fs.open(path,"r+") as f:
                f.truncate(attr.st_size)
        return paramiko.SFTP_OK

    def readlink(self, path):
        return paramiko.SFTP_OP_UNSUPPORTED

    def symlink(self, path):
        return paramiko.SFTP_OP_UNSUPPORTED


class SFTPHandle(paramiko.SFTPHandle):
    """SFTP file handler pointing to a file in an FS object.

    This is a simple file wrapper for SFTPServerInterface, passing read
    and write requests directly through the to underlying file from the FS.
    """

    def __init__(self, owner, path, flags):
        super(SFTPHandle, self).__init__(flags)
        mode = flags_to_mode(flags)
        self.owner = owner
        if not isinstance(path, unicode):
            path = path.decode(self.owner.encoding)
        self.path = path
        self._file = owner.fs.open(path, mode)

    @report_sftp_errors
    def close(self):
        self._file.close()
        return paramiko.SFTP_OK

    @report_sftp_errors
    def read(self, offset, length):
        self._file.seek(offset)
        return self._file.read(length)

    @report_sftp_errors
    def write(self, offset, data):
        self._file.seek(offset)
        self._file.write(data)
        return paramiko.SFTP_OK

    def stat(self):
        return self.owner.stat(self.path)

    def chattr(self,attr):
        return self.owner.chattr(self.path, attr)


class SFTPServer(paramiko.SFTPServer):
    """
    An SFTPServer class that closes the filesystem when done.
    """

    def finish_subsystem(self):
        # Close the SFTPServerInterface, it will close the pyfs file system.
        self.server.close()
        super(SFTPServer, self).finish_subsystem()


class SFTPRequestHandler(SocketServer.BaseRequestHandler):
    """SocketServer RequestHandler subclass for BaseSFTPServer.

    This RequestHandler subclass creates a paramiko Transport, sets up the
    sftp subsystem, and hands off to the transport's own request handling
    thread.
    """
    timeout = 60
    auth_timeout = 60

    def setup(self):
        """
        Creates the SSH transport. Sets security options.
        """
        self.transport = paramiko.Transport(self.request)
        self.transport.load_server_moduli()
        so = self.transport.get_security_options()
        so.digests = ('hmac-sha1', )
        so.compression = ('zlib@openssh.com', 'none')
        self.transport.add_server_key(self.server.host_key)
        self.transport.set_subsystem_handler("sftp", SFTPServer, SFTPServerInterface, self.server.fs, encoding=self.server.encoding)

    def handle(self):
        """
        Start the paramiko server, this will start a thread to handle the connection.
        """
        self.transport.start_server(server=BaseServerInterface())
        # TODO: I like the code below _in theory_ but it does not work as I expected.
        # Figure out how to actually time out a new client if they fail to auth in a
        # certain amount of time.
        #chan = self.transport.accept(self.auth_timeout)
        #if chan is None:
        #    self.transport.close()

    def handle_timeout(self):
        try:
            self.transport.close()
        finally:
            super(SFTPRequestHandler, self).handle_timeout()



class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    pass


class BaseSFTPServer(ThreadedTCPServer):
    """SocketServer.TCPServer subclass exposing an FS via SFTP.

    Operation is in the standard SocketServer style.  The target FS object
    can be passed into the constructor, or set as an attribute on the server::

        server = BaseSFTPServer((hostname,port),fs)
        server.serve_forever()

    It is also possible to specify the host key used by the sever by setting
    the 'host_key' attribute.  If this is not specified, it will default to
    the key found in the DEFAULT_HOST_KEY variable.
    """
    # If the server stops/starts quickly, don't fail because of
    # "port in use" error.
    allow_reuse_address = True

    def __init__(self, address, fs=None, encoding=None, host_key=None, RequestHandlerClass=None):
        self.fs = fs
        self.encoding = encoding
        if host_key is None:
            host_key = DEFAULT_HOST_KEY
        self.host_key = host_key
        if RequestHandlerClass is None:
            RequestHandlerClass = SFTPRequestHandler
        SocketServer.TCPServer.__init__(self, address, RequestHandlerClass)

    def shutdown_request(self, request):
        # Prevent TCPServer from closing the connection prematurely
        return

    def close_request(self, request):
        # Prevent TCPServer from closing the connection prematurely
        return


class BaseServerInterface(paramiko.ServerInterface):
    """
    Paramiko ServerInterface implementation that performs user authentication.

    Note that this base class allows UNAUTHENTICATED ACCESS to the exposed
    FS.  This is intentional, since we can't guess what your authentication
    needs are.  To protect the exposed FS, override the following methods:

        * get_allowed_auths Determine the allowed auth modes
        * check_auth_none Check auth with no credentials
        * check_auth_password Check auth with a password
        * check_auth_publickey Check auth with a public key
    """

    def check_channel_request(self, kind, chanid):
        if kind == 'session':
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_auth_none(self, username):
        """Check whether the user can proceed without authentication."""
        return paramiko.AUTH_SUCCESSFUL

    def check_auth_publickey(self, username,key):
        """Check whether the given public key is valid for authentication."""
        return paramiko.AUTH_FAILED

    def check_auth_password(self, username, password):
        """Check whether the given password is valid for authentication."""
        return paramiko.AUTH_FAILED

    def get_allowed_auths(self,username):
        """Return string containing a comma separated list of allowed auth modes.

        The available modes are  "node", "password" and "publickey".
        """
        return "none"


#  When called from the command-line, expose a TempFS for testing purposes
if __name__ == "__main__":
    from fs.tempfs import TempFS
    server = BaseSFTPServer(("localhost",8022),TempFS())
    try:
        #import rpdb2; rpdb2.start_embedded_debugger('password')
        server.serve_forever()
    except (SystemExit,KeyboardInterrupt):
        server.server_close()




import os
import glob
import json

from traits.trait_base import get_resource_path

import logging
log = logging.getLogger(__name__)

template_subdirs = ["", "../../omnivore8bit/templates"]
try:
    import atrcopy
except ImportError:
    pass
else:
    try:
        path = atrcopy.get_template_path()
        template_subdirs.append(path)
    except AttributeError:
        # skip an old version of atrcopy that happens to be installed
        pass


def construct_path(template_dir, name):
    name = name.lstrip("/")
    if not os.path.isabs(template_dir):
        # resource path will point to the omnivore/templates directory
        path = get_resource_path(1)
        log.debug("resource path: %s" % path)
        pathname = os.path.normpath(os.path.join(path, template_dir, name))
    else:
        pathname = os.path.normpath(os.path.join(template_dir, name))
    return pathname


def find_template_path(name):
    checked = []
    for toplevel in template_subdirs:
        pathname = construct_path(toplevel, name)
        log.debug("Checking for template at %s" % pathname)
        if os.path.exists(pathname):
            log.debug("Found template for %s: %s" % (name, pathname))
            return pathname
        checked.append(os.path.abspath(toplevel))
    else:
        raise OSError("No template found for %s in %s" % (name, str(checked)))


def get_template_fh(name):
    # resource path will point to the omnivore/templates directory
    pathname = find_template_path(name)
    log.debug("Loading template for %s: %s" % (name, pathname))
    fh = open(pathname, "rb")
    return fh


def get_template(name):
    # resource path will point to the omnivore/templates directory
    fh = get_template_fh(name)
    source = fh.read()
    return source


def iter_templates():
    for toplevel in template_subdirs:
        pathname = construct_path(toplevel, "*")
        for template in glob.glob(pathname):
            inf = template + ".inf"
            if os.path.exists(template) and os.path.isfile(template) and os.path.exists(inf):
                try:
                    with open(inf, "r") as fh:
                        j = json.loads(fh.read())
                except ValueError:
                    j = {}
                    raise
                j["pathname"] = template
                j["uri"] = "template://" + os.path.basename(template)
                log.debug("template json: %s" % str(template))
                yield j



# All built-in recognizer plugins should be listed in this file so that the
# application can import this single file and determine the default plugins.
# In addition, a list of plugins is expected so that the framework can import
# all built-in plugins at once.
#
# A cog script is included to automatically generate the expected code. Test with::
#
#    cog.py __init__.py
#
# which prints to standard output. To place the generated code in the file, use::
#
#    cog.py -r __init__.py
#
# Note that the cog script works when called from the top directory (.i.e.  at
# the same level as the omnivore directory) or this directory.

# [[[cog
#import os
#import sys
#import inspect
#import imp
#
#from envisage.api import Plugin
#
#cwd = os.getcwd()
#cog.msg("working dir : %s" % cwd)
#path = os.path.dirname(os.path.join(cwd, cog.inFile))
#cog.msg("scanning dir: %s" % path)
#top = os.path.abspath(os.path.join(path, "../../..")) # so absolute imports of omnivore will work
#sys.path.append(top)
#cog.msg("top dir     : %s" % top)
#import glob
#cog.outl("plugins = []")
#for filename in glob.iglob(os.path.join(path, "*.py")):
#    if filename.endswith("__init__.py"):
#        continue
#    cog.msg("filename: %s" % filename)
#    modname = filename.split(".py")[0].split("/")[-1]
#    module = imp.load_source(modname, filename)
#    cog.msg("module: %s" % module.__name__)
#    members = inspect.getmembers(module, inspect.isclass)
#    names = []
#    for name, cls in members:
#        cog.msg("member: %s, cls %s, module %s" % (name, cls, cls.__module__))
#        if issubclass(cls, Plugin) and cls.__module__ == module.__name__:
#            names.append(name)
#    if names:
#       cog.outl("from %s import %s" % (modname, ", ".join(names)))
#       for name in names:
#           cog.outl("plugins.append(%s())" % name)
# ]]]*/
plugins = []
from exception_handler import ExceptionHandlerPlugin
plugins.append(ExceptionHandlerPlugin())
from file_progress import FileProgressPlugin
plugins.append(FileProgressPlugin())
from open_recent import OpenRecentPlugin
plugins.append(OpenRecentPlugin())
#[[[end]]]



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/22/2007
#
#-------------------------------------------------------------------------

""" Traits UI simple, themed slider-based integer or float value editor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from math \
    import log10, pow

from traits.api \
    import HasPrivateTraits, Instance, Enum, Range, Str, Float, Bool, Color, \
    TraitError

from traitsui.ui_traits \
    import Alignment

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

from pyface.timer.api \
    import do_after

from constants \
    import ErrorColor

from helper \
    import disconnect, disconnect_no_id, BufferDC

#-------------------------------------------------------------------------
#  '_ThemedSliderEditor' class:
#-------------------------------------------------------------------------


class _ThemedSliderEditor(Editor):
    """ Traits UI simple, themed slider-based integer or float value editor.
    """

    # The low end of the slider range:
    low = Float

    # The high end of the slider range:
    high = Float

    # The smallest allowed increment:
    increment = Float

    # The current text being displayed:
    text = Str

    #-- Class Variables ------------------------------------------------------

    text_styles = {
        'left': wx.TE_LEFT,
        'center': wx.TE_CENTRE,
        'right': wx.TE_RIGHT
    }

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        # Establish the range of the slider:
        low, high = factory.low, factory.high
        if high <= low:
            low = high = None
            range = self.object.trait(self.name).handler
            if isinstance(range, Range):
                low, high = range._low, range._high
            if low is None:
                if high is None:
                    high = 1.0
                low = high - 1.0
            elif high is None:
                high = low + 1.0

        # Establish the slider increment:
        increment = factory.increment
        if increment <= 0.0:
            if isinstance(low, int):
                increment = 1.0
            else:
                increment = pow(10, int(log10((high - low) / 1000.0)))

        # Save the values we calculated:
        self.set(low=low, high=high, increment=increment)

        # Create the control:
        self.control = control = wx.Window(parent, -1,
                                           size=wx.Size(70, 20),
                                           style=wx.FULL_REPAINT_ON_RESIZE |
                                           wx.TAB_TRAVERSAL)

        # Set up the painting event handlers:
        wx.EVT_ERASE_BACKGROUND(control, self._erase_background)
        wx.EVT_PAINT(control, self._on_paint)
        wx.EVT_SET_FOCUS(control, self._set_focus)

        # Set up mouse event handlers:
        wx.EVT_LEFT_DOWN(control, self._left_down)
        wx.EVT_LEFT_UP(control, self._left_up)
        wx.EVT_MOTION(control, self._motion)
        wx.EVT_MOUSEWHEEL(control, self._mouse_wheel)
        wx.EVT_ENTER_WINDOW(control, self._enter_window)
        wx.EVT_LEAVE_WINDOW(control, self._leave_window)

        # Set up the control resize handler:
        wx.EVT_SIZE(control, self._resize)

        # Set the tooltip:
        if not self.set_tooltip():
            control.SetToolTipString('[%g..%g]' % (low, high))

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect_no_id(
            self.control,
            wx.EVT_ERASE_BACKGROUND,
            wx.EVT_PAINT,
            wx.EVT_SET_FOCUS,
            wx.EVT_LEFT_DOWN,
            wx.EVT_LEFT_UP,
            wx.EVT_MOTION,
            wx.EVT_MOUSEWHEEL,
            wx.EVT_ENTER_WINDOW,
            wx.EVT_LEAVE_WINDOW,
            wx.EVT_SIZE)

        if self._text is not None:
            disconnect(self._text, wx.EVT_TEXT_ENTER)
            disconnect_no_id(
                self._text,
                wx.EVT_KILL_FOCUS,
                wx.EVT_ENTER_WINDOW,
                wx.EVT_LEAVE_WINDOW,
                wx.EVT_CHAR)

        super(_ThemedSliderEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.text = '%g' % self.value
        self._text_size = None
        self._refresh()

        if self._text is not None:
            self._text.SetValue(self.text)

    #-------------------------------------------------------------------------
    #  Updates the object when the control slider value changes:
    #-------------------------------------------------------------------------

    def update_object(self, value):
        """ Updates the object when the control slider value changes.
        """
        try:
            self.value = value
        except TraitError:
            self.value = int(value)

        self.update_editor()

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        pass

    #-- Private Methods ------------------------------------------------------

    def _get_text_bounds(self):
        """ Get the window bounds of where the current text should be
            displayed.
        """
        tdx, tdy, descent, leading = self._get_text_size()
        wdx, wdy = self.control.GetClientSizeTuple()
        ty = ((wdy - (tdy - descent)) / 2) - 2
        alignment = self.factory.alignment
        if alignment == 'left':
            tx = 4
        elif alignment == 'center':
            tx = (wdx - tdx) / 2
        else:
            tx = wdx - tdx - 4

        return (tx, ty, tdx, tdy)

    def _get_text_size(self):
        """ Returns the text size information for the window.
        """
        if self._text_size is None:
            self._text_size = self.control.GetFullTextExtent(
                self.text.strip() or 'M')

        return self._text_size

    def _refresh(self):
        """ Refreshes the contents of the control.
        """
        if self.control is not None:
            self.control.Refresh()

    def _set_slider_position(self, x):
        """ Calculates a new slider value for a specified (x,y) coordinate.
        """
        wdx, wdy = self.control.GetSizeTuple()
        if 3 <= x < wdx:
            value = self.low + (((x - 3) * (self.high - self.low)) / (wdx - 4))
            increment = self.increment
            if increment > 0:
                value = round(value / increment) * increment
            self.update_object(value)

    def _delayed_click(self):
        """ Handle a delayed click response.
        """
        if self._pending:
            self._pending = False
            self._set_slider_position(self._x)

    def _pop_up_text(self):
        """ Pop-up a text control to allow the user to enter a value using
            the keyboard.
        """
        control = self.control
        self._text = text = wx.TextCtrl(control, -1, self.text,
                                        size=control.GetSize(),
                                        style=self.text_styles[self.factory.alignment] |
                                        wx.TE_PROCESS_ENTER)
        text.SetSelection(-1, -1)
        text.SetFocus()
        wx.EVT_TEXT_ENTER(control, text.GetId(), self._text_completed)
        wx.EVT_KILL_FOCUS(text, self._text_completed)
        wx.EVT_ENTER_WINDOW(text, self._enter_text)
        wx.EVT_LEAVE_WINDOW(text, self._leave_text)
        wx.EVT_CHAR(text, self._key_entered)

    def _destroy_text(self):
        """ Destroys the current text control.
        """
        self._ignore_focus = self._in_text_window
        disconnect(self._text, wx.EVT_TEXT_ENTER)
        disconnect_no_id(self._text, wx.EVT_KILL_FOCUS, wx.EVT_ENTER_WINDOW,
                         wx.EVT_LEAVE_WINDOW, wx.EVT_CHAR)
        self.control.DestroyChildren()
        self._text = None

    #--- wxPython Event Handlers ---------------------------------------------

    def _erase_background(self, event):
        """ Do not erase the background here (do it in the 'on_paint' handler).
        """
        pass

    def _on_paint(self, event):
        """ Paint the background using the associated ImageSlice object.
        """
        control = self.control
        dc = BufferDC(control)

        # Draw the slider bar:
        wdx, wdy = control.GetClientSize()
        dx = max(0, min(wdx - 2,
                        int(round(((wdx - 3) * (self.value - self.low)) /
                                  (self.high - self.low)))))

        factory = self.factory
        dc.SetBrush(wx.Brush(factory.slider_color_))
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(0, 0, dx + 3, wdy)

        # Draw the rest of the background:
        dc.SetBrush(wx.Brush(factory.bg_color_))
        dc.DrawRectangle(dx + 3, 0, wdx - dx - 3, wdy)

        # Draw the slider tip:
        dc.SetBrush(wx.Brush(factory.tip_color_))
        dc.DrawRectangle(dx, 0, 3, wdy)

        # Draw the current text value (if requested):
        if factory.show_value:
            dc.SetBackgroundMode(wx.TRANSPARENT)
            dc.SetTextForeground(factory.text_color_)
            dc.SetFont(control.GetFont())
            tx, ty, tdx, tdy = self._get_text_bounds()
            dc.DrawText(self.text, tx, ty)

        # Copy the buffer to the display:
        dc.copy()

    def _resize(self, event):
        """ Handles the control being resized.
        """
        if self._text is not None:
            self._text.SetSize(self.control.GetSize())

    def _set_focus(self, event):
        """ Handle the control getting the keyboard focus.
        """
        if ((not self._ignore_focus) and
            (self._x is None) and
                (self._text is None)):
            self._pop_up_text()

        event.Skip()

    def _left_down(self, event):
        """ Handles the left mouse being pressed.
        """
        self._x, self._y = event.GetX(), event.GetY()
        self._pending = True
        self.control.CaptureMouse()
        do_after(150, self._delayed_click)

    def _left_up(self, event):
        """ Handles the left mouse button being released.
        """
        if self._x is not None:
            self.control.ReleaseMouse()

        if self._pending:
            self._pop_up_text()

        self._x = self._y = self._pending = None

    def _motion(self, event):
        """ Handles the mouse moving.
        """
        if self._x is not None:
            x, y = event.GetX(), event.GetY()
            if self._pending:
                if (abs(x - self._x) + abs(y - self._y)) < 3:
                    return
                self._pending = False
            self._set_slider_position(x)

    def _mouse_wheel(self, event):
        """ Handles the mouse wheel rotating.
        """
        if self._in_window:
            increment = event.GetWheelRotation() / event.GetWheelDelta()
            delta = (self.high - self.low) / 100.0
            if isinstance(self.value, int) and (abs(delta) < 1):
                delta = int(abs(delta) / delta)

            self.update_object(min(max(self.value + increment * delta,
                                       self.low), self.high))

    def _enter_window(self, event):
        """ Handles the mouse pointer entering the control.
        """
        self._in_window = True

        if not self._ignore_focus:
            self._ignore_focus = True
            self.control.SetFocus()

        self._ignore_focus = False

    def _leave_window(self, event):
        """ Handles the mouse pointer leaving the control.
        """
        self._in_window = False

    def _update_value(self, event):
        """ Updates the object value from the current text control value.
        """
        control = event.GetEventObject()
        try:
            self.update_object(float(control.GetValue()))

            return True

        except TraitError:
            control.SetBackgroundColour(ErrorColor)
            control.Refresh()

            return False

    def _enter_text(self, event):
        """ Handles the mouse entering the pop-up text control.
        """
        self._in_text_window = True

    def _leave_text(self, event):
        """ Handles the mouse leaving the pop-up text control.
        """
        self._in_text_window = False

    def _text_completed(self, event):
        """ Handles the user pressing the 'Enter' key in the text control.
        """
        if isinstance(event, wx.FocusEvent):
            event.Skip()
        if self._update_value(event):
            self._destroy_text()

    def _key_entered(self, event):
        """ Handles individual key strokes while the text control is active.
        """
        key_code = event.GetKeyCode()
        if key_code == wx.WXK_ESCAPE:
            self._destroy_text()
            return

        if key_code == wx.WXK_TAB:
            if self._update_value(event):
                if event.ShiftDown():
                    self.control.Navigate(0)
                else:
                    self.control.Navigate()
            return

        event.Skip()

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# wxPython editor factory for themed slider editors:


class ThemedSliderEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = _ThemedSliderEditor

    # The low end of the slider range:
    low = Float

    # The high end of the slider range:
    high = Float

    # The smallest allowed increment:
    increment = Float

    # Should the current value be displayed as text?
    show_value = Bool(True)

    # The alignment of the text within the slider:
    alignment = Alignment('center')

    # The color to use for the slider bar:
    slider_color = Color(0xC0C0C0)

    # The background color for the slider:
    bg_color = Color('white')

    # The color of the slider tip:
    tip_color = Color(0xFF7300)

    # The color to use for the value text:
    text_color = Color('black')



""" This module implements functionality similar to class decorators.

At the moment it is used for the 'implements' and 'adapts' functions, which
have been deprecated together with the 'traits.protocols' package.

Please don't import anything from this module!

"""

from __future__ import absolute_import

from types import FunctionType

import sys

from .. import _py2to3

__all__ = ['addClassAdvisor']


def metamethod(func):
    """Wrapper for metaclass method that might be confused w/instance method"""
    return property(lambda ob: func.__get__(ob,ob.__class__))


if sys.version_info[0] < 3:
    from types import ClassType, InstanceType
    
    ClassicTypes = ClassType
    
    def classicMRO(ob, extendedClassic=False):
        stack = []
        push = stack.insert
        pop = stack.pop
        push(0,ob)
        while stack:
            cls = pop()
            yield cls
            p = len(stack)
            for b in cls.__bases__: push(p,b)
        if extendedClassic:
            yield InstanceType
            yield object
    
    
    def getMRO(ob, extendedClassic=False):
    
        if isinstance(ob,ClassicTypes):
            return classicMRO(ob,extendedClassic)
    
        elif isinstance(ob,type):
            return ob.__mro__
    
        return ob,
else:
    def getMRO(ob, *args, **kwargs):
        if args or kwargs:
            kwargs.pop('extendedClassic',None)
            if len(args)>1 or kwargs:
                raise TypeError
            import warnings
            warnings.warn(DeprecationWarning(
                """In Python 3 there are no more ols-style classes.
                Therefore, extendedClassic has no meaning and should not be used.
                """
            ))
        if isinstance(ob,type):
            return ob.__mro__
    
        return ob,

try:
    from ._speedups import metamethod, getMRO, classicMRO
except ImportError:
    pass


# property-safe 'super()' for Python 2.2; 2.3 can use super() instead

def supermeta(typ,ob):

    starttype = type(ob)
    mro = starttype.__mro__
    if typ not in mro:
        starttype = ob
        mro = starttype.__mro__

    mro = iter(mro)
    for cls in mro:
        if cls is typ:
            mro = [cls.__dict__ for cls in mro]
            break
    else:
        raise TypeError("Not sub/supertypes:", starttype, typ)

    typ = type(ob)

    class theSuper(object):

        def __getattribute__(self,name):
            for d in mro:
                if name in d:
                    descr = d[name]
                    try:
                        descr = descr.__get__
                    except AttributeError:
                        return descr
                    else:
                        return descr(ob,typ)
            return object.__getattribute__(self,name)

    return theSuper()


def getFrameInfo(frame):
    """Return (kind,module,locals,globals) for a frame

    'kind' is one of "exec", "module", "class", "function call", or "unknown".
    """

    f_locals = frame.f_locals
    f_globals = frame.f_globals

    sameNamespace = f_locals is f_globals
    hasModule = '__module__' in f_locals
    hasName = '__name__' in f_globals

    sameName = hasModule and hasName
    sameName = sameName and f_globals['__name__']==f_locals['__module__']

    module = hasName and sys.modules.get(f_globals['__name__']) or None

    namespaceIsModule = module and module.__dict__ is f_globals

    if not namespaceIsModule:
        # some kind of funky exec
        kind = "exec"
    elif sameNamespace and not hasModule:
        kind = "module"
    elif sameName and not sameNamespace:
        kind = "class"
    elif not sameNamespace:
        kind = "function call"
    else:
        # How can you have f_locals is f_globals, and have '__module__' set?
        # This is probably module-level code, but with a '__module__' variable.
        kind = "unknown"

    return kind,module,f_locals,f_globals


def addClassAdvisor(callback, depth=2):

    """Set up 'callback' to be passed the containing class upon creation

    This function is designed to be called by an "advising" function executed
    in a class suite.  The "advising" function supplies a callback that it
    wishes to have executed when the containing class is created.  The
    callback will be given one argument: the newly created containing class.
    The return value of the callback will be used in place of the class, so
    the callback should return the input if it does not wish to replace the
    class.

    The optional 'depth' argument to this function determines the number of
    frames between this function and the targeted class suite.  'depth'
    defaults to 2, since this skips this function's frame and one calling
    function frame.  If you use this function from a function called directly
    in the class suite, the default will be correct, otherwise you will need
    to determine the correct depth yourself.

    This function works by installing a special class factory function in
    place of the '__metaclass__' of the containing class.  Therefore, only
    callbacks *after* the last '__metaclass__' assignment in the containing
    class will be executed.  Be sure that classes using "advising" functions
    declare any '__metaclass__' *first*, to ensure all callbacks are run.

    Moreover, since starting from Python 3, metaclasses are specified
    differently, this function does not work anymore. Worse, as the metaclass
    is selected even before running the class's body, there is no way to
    fix this in a general way. As long as the metaclass provides some hooks
    to run code at class creation time, we can use them, but standard "type"
    does not.    
    """
    if sys.version_info[0] >= 3:
        raise NotImplementedError("Class advisors are not possible in python 3.")


    frame = sys._getframe(depth)
    kind, module, caller_locals, caller_globals = getFrameInfo(frame)

    if kind not in ("class", "exec"):
        raise SyntaxError(
            "Advice must be in the body of a class statement"
        )

    previousMetaclass = caller_locals.get('__metaclass__')
    defaultMetaclass  = caller_globals.get('__metaclass__', type)  #TODO: This used to be ClassType, but I think this was errornous. Check it!


    def advise(name,bases,cdict):

        if '__metaclass__' in cdict:
            del cdict['__metaclass__']

        if previousMetaclass is None:
             if bases:
                 # find best metaclass or use global __metaclass__ if no bases
                 meta = determineMetaclass(bases)
             else:
                 meta = defaultMetaclass

        elif isClassAdvisor(previousMetaclass):
            # special case: we can't compute the "true" metaclass here,
            # so we need to invoke the previous metaclass and let it
            # figure it out for us (and apply its own advice in the process)
            meta = previousMetaclass

        else:
            meta = determineMetaclass(bases, previousMetaclass)

        newClass = meta(name,bases,cdict)

        # this lets the callback replace the class completely, if it wants to
        return callback(newClass)

    # introspection data only, not used by inner function
    advise.previousMetaclass = previousMetaclass
    advise.callback = callback

    # install the advisor
    caller_locals['__metaclass__'] = advise


def isClassAdvisor(ob):
    """True if 'ob' is a class advisor function"""
    return isinstance(ob,FunctionType) and hasattr(ob,'previousMetaclass')


def determineMetaclass(bases, explicit_mc=None):

    """Determine metaclass from 1+ bases and optional explicit __metaclass__"""

    meta = [getattr(b,'__class__',type(b)) for b in bases]

    if explicit_mc is not None:
        # The explicit metaclass needs to be verified for compatibility
        # as well, and allowed to resolve the incompatible bases, if any
        meta.append(explicit_mc)

    if len(meta)==1:
        # easy case
        return meta[0]

    candidates = minimalBases(meta) # minimal set of metaclasses

    if not candidates:
        # they're all "classic" classes
        # should never happen in Python 3, so this should be fine
        from types import ClassType
        return ClassType

    elif len(candidates)>1:
        # We could auto-combine, but for now we won't...
        raise TypeError("Incompatible metatypes",bases)

    # Just one, return it
    return candidates[0]


def minimalBases(classes):
    """Reduce a list of base classes to its ordered minimum equivalent"""

    classes = [c for c in classes if not _py2to3.is_old_style_class(c)]
    candidates = []

    for m in classes:
        for n in classes:
            if issubclass(n,m) and m is not n:
                break
        else:
            # m has no subclasses in 'classes'
            if m in candidates:
                candidates.remove(m)    # ensure that we're later in the list
            candidates.append(m)

    return candidates


from weakref import ref

class StrongRef(object):

    """Like a weakref, but for non-weakrefable objects"""

    __slots__ = 'referent'

    def __init__(self,referent):
        self.referent = referent

    def __call__(self):
        return self.referent

    def __hash__(self):
        return hash(self.referent)

    def __eq__(self,other):
        return self.referent==other

    def __repr__(self):
        return 'StrongRef(%r)' % self.referent


def mkRef(ob,*args):
    """Return either a weakref or a StrongRef for 'ob'

    Note that extra args are forwarded to weakref.ref() if applicable."""

    try:
        return ref(ob,*args)
    except TypeError:
        return StrongRef(ob)




#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The node manager looks after a collection of node types. """

# Standard library imports.
import logging

# Enthought library imports.
from traits.api import HasPrivateTraits, List

# Local imports
from .node_type import NodeType


# Create a logger for this module.
logger = logging.getLogger(__name__)


class NodeManager(HasPrivateTraits):
    """ The node manager looks after a collection of node types. """

    #### 'NodeManager' interface ##########################################

    # All registered node types.
    node_types = List(NodeType)

    # fixme: Where should the system actions go?  The node tree, the node
    # tree model, here?!?
    system_actions = List

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Creates a new tree model. """

        # Base class constructor.
        super(NodeManager, self).__init__(**traits)

        # This saves looking up a node's type every time.  If we ever have
        # nodes that change type dynamically then we will obviously have to
        # re-think this (although we should probably re-think dynamic type
        # changes first ;^).
        self._node_to_type_map = {} # { Any node : NodeType node_type }

        return

    ###########################################################################
    # 'NodeManager' interface.
    ###########################################################################

    # fixme: This is the only API call that we currently have that manipulates
    # the manager's node types.  Should we make the 'node_types' list
    # available via the public API?
    def add_node_type(self, node_type):
        """ Adds a new node type to the manager. """

        node_type.node_manager = self
        self.node_types.append(node_type)

        return

    def get_node_type(self, node):
        """ Returns the node's type.

        Returns None if none of the manager's node types recognize the node.

        """

        # Generate the key for the node to type map.
        key = self.get_key(node)

        # Check the cache first.
        node_type = self._node_to_type_map.get(key, None)
        if node_type is None:
            # If we haven't seen this node before then attempt to find a node
            # type that 'recognizes' it.
            #
            # fixme: We currently take the first node type that recognizes the
            # node.  This obviously means that ordering of node types is
            # important,  but we don't have an interface for controlling the
            # order.  Maybe sort on some 'precedence' trait on the node type?
            for node_type in self.node_types:
                if node_type.is_type_for(node):
                    self._node_to_type_map[key] = node_type
                    break

            else:
                node_type = None

        if node_type is None:
            logger.warn('no node type for %s' % str(node))

        return node_type

    def get_key(self, node):
        """ Generates a unique key for a node.

        In this case,  'unique' means unqiue within the node manager.

        """

        # We do it like this 'cos, for example, using id() on a string doesn't
        # give us what we want, but things like lists aren't hashable, so we
        # can't always use hash()).
        try:
            key = hash(node)

        except:
            key = id(node)

        return key

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _node_types_changed(self, new):
        """ Called when the entire list of node types has been changed. """

        for node_type in new:
            node_type.node_manager = self

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Defines the Item class, which is used to represent a single item within
    a Traits-based user interface.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import re

from traits.api import (Bool, Callable, Constant, Delegate, Float, Instance,
                        Range, Str, Undefined, Dict,)

from traits.trait_base import user_name_for
import traits._py2to3 as _py2to3

from .view_element import ViewSubElement

from .ui_traits import convert_theme, ContainerDelegate, EditorStyle

from .editor_factory import EditorFactory

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Pattern of all digits:
all_digits = re.compile(r'\d+')

# Pattern for finding size infomation embedded in an item description:
size_pat = re.compile(r"^(.*)<(.*)>(.*)$", re.MULTILINE | re.DOTALL)

# Pattern for finding tooltip infomation embedded in an item description:
tooltip_pat = re.compile(r"^(.*)`(.*)`(.*)$", re.MULTILINE | re.DOTALL)

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Reference to an EditorFactory:
ItemEditor = Instance(EditorFactory, allow_none=True)

# Amount of padding to add around an item:
Padding = Range(-15, 15, 0, desc='amount of padding to add around item')

#-------------------------------------------------------------------------
#  'Item' class:
#-------------------------------------------------------------------------


class Item(ViewSubElement):
    """ An element in a Traits-based user interface.

    Magic:

    - Items are rendered as layout elements if :attr:`name` is set to
      special values:

      * name='', the item is rendered as a static label

      * name='_', the item is rendered as a separator

      * name=' ', the item is rendered as a 5 pixel spacer

      * name='23' (any number), the item is rendered as a spacer of
        the size specified (number of pixels)
    """

    # FIXME: all the logic for the name = '', '_', ' ', '23' magic is in
    # _GroupPanel._add_items in qt/ui_panel.py, which is a very unlikely place
    # to look for it. Ideally, that logic should be in this class.

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # A unique identifier for the item. If not set, it defaults to the value
    # of **name**.
    id = Str

    # User interface label for the item in the GUI. If this attribute is not
    # set, the label is the value of **name** with slight modifications:
    # underscores are replaced by spaces, and the first letter is capitalized.
    # If an item's **name** is not specified, its label is displayed as
    # static text, without any editor widget.
    label = Str

    # Name of the trait the item is editing:
    name = Str

    # Style-sheet to apply to item / group (Qt only)
    style_sheet = Str

    # Help text describing the purpose of the item. The built-in help handler
    # displays this text in a pop-up window if the user clicks the widget's
    # label. View-level help displays the help text for all items in a view.
    # If this attribute is not set, the built-in help handler generates a
    # description based on the trait definition.
    help = Str

    # The HasTraits object whose trait attribute the item is editing:
    object = ContainerDelegate

    # Presentation style for the item:
    style = ContainerDelegate

    # Docking style for the item:
    dock = ContainerDelegate

    # Image to display on notebook tabs:
    image = ContainerDelegate

    # The theme to use for the item itself:
    item_theme = ContainerDelegate

    # The theme to use for the item's label:
    label_theme = ContainerDelegate

    # Category of elements dragged from view:
    export = ContainerDelegate

    # Should a label be displayed for the item?
    show_label = Delegate('container', 'show_labels')

    # Editor to use for the item:
    editor = ItemEditor

    # Additional editor traits to be set if default traits editor to be used:
    editor_args = Dict

    # Should the item use extra space along its Group's non-layout axis? If set to
    # True, the widget expands to fill any extra space that is available in the
    # display. If set to True for more than one item in the same View, any extra
    # space is divided between them. If set to False, the widget uses only
    # whatever space it is explicitly (or implicitly) assigned. The default
    # value of Undefined means that the use (or non-use) of extra space will be
    # determined by the editor associated with the item.
    resizable = Bool(Undefined)

    # Should the item use extra space along its Group's layout axis? For
    # example, it a vertical group, should an item expand vertically to use
    # any extra space available in the group?
    springy = Bool(False)

    # Should the item use any extra space along its Group's non-layout
    # orientation? For example, in a vertical group, should an item expand
    # horizontally to the full width of the group? If left to the default value
    # of Undefined, the decision will be left up to the associated item editor.
    full_size = Bool(Undefined)

    # Should the item's label use emphasized text? If the label is not shown,
    # this attribute is ignored.
    emphasized = Bool(False)

    # Should the item receive focus initially?
    has_focus = Bool(False)

    # Pre-condition for including the item in the display. If the expression
    # evaluates to False, the item is not defined in the display. Conditions
    # for **defined_when** are evaluated only once, when the display is first
    # constructed. Use this attribute for conditions based on attributes that
    # vary from object to object, but that do not change over time. For example,
    # displaying a 'maiden_name' item only for female employees in a company
    # database.
    defined_when = Str

    # Pre-condition for showing the item. If the expression evaluates to False,
    # the widget is not visible (and disappears if it was previously visible).
    # If the value evaluates to True, the widget becomes visible. All
    # **visible_when** conditions are checked each time that any trait value
    # is edited in the display. Therefore, you can use **visible_when**
    # conditions to hide or show widgets in response to user input.
    visible_when = Str

    # Pre-condition for enabling the item. If the expression evaluates to False,
    # the widget is disabled, that is, it does not accept input. All
    # **enabled_when** conditions are checked each time that any trait value
    # is edited in the display. Therefore, you can use **enabled_when**
    # conditions to enable or disable widgets in response to user input.
    enabled_when = Str

    # Amount of extra space, in pixels, to add around the item. Values must be
    # integers between -15 and 15. Use negative values to subtract from the
    # default spacing.
    padding = Padding

    # Tooltip to display over the item, when the mouse pointer is left idle
    # over the widget. Make this text as concise as possible; use the **help**
    # attribute to provide more detailed information.
    tooltip = Str

    # A Callable to use for formatting the contents of the item. This function
    # or method is called to create the string representation of the trait value
    # to be edited. If the widget does not use a string representation, this
    # attribute is ignored.
    format_func = Callable

    # Python format string to use for formatting the contents of the item.
    # The format string is applied to the string representation of the trait
    # value before it is displayed in the widget. This attribute is ignored if
    # the widget does not use a string representation, or if the
    # **format_func** is set.
    format_str = Str

    # Requested width of the editor (in pixels or fraction of available width).
    # For pixel values (i.e. values not in the range from 0.0 to 1.0), the
    # actual displayed width is at least the maximum of **width** and the
    # optimal width of the widget as calculated by the GUI toolkit. Specify a
    # negative value to ignore the toolkit's optimal width. For example, use
    # -50 to force a width of 50 pixels. The default value of -1 ensures that
    # the toolkit's optimal width is used.
    #
    # A value in the range from 0.0 to 1.0 specifies the fraction of the
    # available width to assign to the editor. Note that the value is not an
    # absolute value, but is relative to other item's whose **width** is also
    # in the 0.0 to 1.0 range. For example, if you have two item's with a width
    # of 0.1, and one item with a width of 0.2, the first two items will each
    # receive 25% of the available width, while the third item will receive
    # 50% of the available width. The available width is the total width of the
    # view minus the width of any item's with fixed pixel sizes (i.e. width
    # values not in the 0.0 to 1.0 range).
    width = Float(-1.0)

    # Requested height of the editor (in pixels or fraction of available
    # height). For pixel values (i.e. values not in the range from 0.0 to 1.0),
    # the actual displayed height is at least the maximum of **height** and the
    # optimal height of the widget as calculated by the GUI toolkit. Specify a
    # negative value to ignore the toolkit's optimal height. For example, use
    # -50 to force a height of 50 pixels. The default value of -1 ensures that
    # the toolkit's optimal height is used.
    #
    # A value in the range from 0.0 to 1.0 specifies the fraction of the
    # available height to assign to the editor. Note that the value is not an
    # absolute value, but is relative to other item's whose **height** is also
    # in the 0.0 to 1.0 range. For example, if you have two item's with a height
    # of 0.1, and one item with a height of 0.2, the first two items will each
    # receive 25% of the available height, while the third item will receive
    # 50% of the available height. The available height is the total height of
    # the view minus the height of any item's with fixed pixel sizes (i.e.
    # height values not in the 0.0 to 1.0 range).
    height = Float(-1.0)

    # The extended trait name of the trait containing the item's invalid state
    # status (passed through to the item's editor):
    invalid = Str

    #-------------------------------------------------------------------------
    #  Initialize the object:
    #-------------------------------------------------------------------------

    def __init__(self, value=None, **traits):
        """ Initializes the item object.
        """
        super(Item, self).__init__(**traits)

        if value is None:
            return

        if not isinstance(value, basestring):
            raise TypeError(
                "The argument to Item must be a string of the "
                "form: [id:][object.[object.]*][name]['['label']']`tooltip`"
                "[<width[,height]>][#^][$|@|*|~|;style]")

        value, empty = self._parse_label(value)
        if empty:
            self.show_label = False

        value = self._parse_style(value)
        value = self._parse_size(value)
        value = self._parse_tooltip(value)
        value = self._option(value, '#', 'resizable', True)
        value = self._option(value, '^', 'emphasized', True)
        value = self._split('id', value, ':', _py2to3.str_find, 0, 1)
        value = self._split('object', value, '.', _py2to3.str_rfind, 0, 1)

        if value != '':
            self.name = value

    #-------------------------------------------------------------------------
    #  Returns whether or not the object is replacable by an Include object:
    #-------------------------------------------------------------------------

    def is_includable(self):
        """ Returns a Boolean indicating whether the object is replaceable by an
            Include object.
        """
        return (self.id != '')

    #-------------------------------------------------------------------------
    #  Returns whether or not the Item represents a spacer or separator:
    #-------------------------------------------------------------------------

    def is_spacer(self):
        """ Returns True if the item represents a spacer or separator.
        """
        name = self.name.strip()

        return ((name == '') or (name == '_') or
                (all_digits.match(name) is not None))

    #-------------------------------------------------------------------------
    #  Gets the help text associated with the Item in a specified UI:
    #-------------------------------------------------------------------------

    def get_help(self, ui):
        """ Gets the help text associated with the Item in a specified UI.
        """
        # Return 'None' if the Item is a separator or spacer:
        if self.is_spacer():
            return None

        # Otherwise, it must be a trait Item:
        if self.help != '':
            return self.help

        object = eval(self.object_, globals(), ui.context)

        return object.base_trait(self.name).get_help()

    #-------------------------------------------------------------------------
    #  Gets the label to use for a specified Item in a specified UI:
    #-------------------------------------------------------------------------

    def get_label(self, ui):
        """ Gets the label to use for a specified Item.

        If not specified, the label is set as the name of the
        corresponding trait, replacing '_' with ' ', and capitalizing
        the first letter (see :func:`user_name_for`). This is called
        the *user name*.

        Magic:

        - if attr:`item.label` is specified, and it begins with '...',
          the final label is the user name followed by the item label
        - if attr:`item.label` is specified, and it ends with '...',
          the final label is the item label followed by the user name
        """
        # Return 'None' if the Item is a separator or spacer:
        if self.is_spacer():
            return None

        label = self.label
        if label != '':
            return label

        name = self.name
        object = eval(self.object_, globals(), ui.context)
        trait = object.base_trait(name)
        label = user_name_for(name)
        tlabel = trait.label
        if tlabel is None:
            return label

        if isinstance(tlabel, basestring):
            if tlabel[0:3] == '...':
                return label + tlabel[3:]
            if tlabel[-3:] == '...':
                return tlabel[:-3] + label
            if self.label != '':
                return self.label
            return tlabel

        return tlabel(object, name, label)

    #-------------------------------------------------------------------------
    #  Returns an id used to identify the item:
    #-------------------------------------------------------------------------

    def get_id(self):
        """ Returns an ID used to identify the item.
        """
        if self.id != '':
            return self.id

        return self.name

    #-------------------------------------------------------------------------
    #  Parses a '<width,height>' value from the string definition:
    #-------------------------------------------------------------------------

    def _parse_size(self, value):
        """ Parses a '<width,height>' value from the string definition.
        """
        match = size_pat.match(value)
        if match is not None:
            data = match.group(2)
            value = match.group(1) + match.group(3)
            col = data.find(',')
            if col < 0:
                self._set_float('width', data)
            else:
                self._set_float('width', data[: col])
                self._set_float('height', data[col + 1:])

        return value

    #-------------------------------------------------------------------------
    #  Parses a '`tooltip`' value from the string definition:
    #-------------------------------------------------------------------------

    def _parse_tooltip(self, value):
        """ Parses a *tooltip* value from the string definition.
        """
        match = tooltip_pat.match(value)
        if match is not None:
            self.tooltip = match.group(2)
            value = match.group(1) + match.group(3)

        return value

    #-------------------------------------------------------------------------
    #  Sets a specified trait to a specified string converted to a float:
    #-------------------------------------------------------------------------

    def _set_float(self, name, value):
        """ Sets a specified trait to a specified string converted to a float.
        """
        value = value.strip()
        if value != '':
            setattr(self, name, float(value))

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of the Item:
    #-------------------------------------------------------------------------

    def __repr__(self):
        """ Returns a "pretty print" version of the Item.
        """

        options = self._repr_options('id', 'object', 'label', 'style',
                                     'show_label', 'width', 'height')
        if options is None:
            return "Item( '%s' )" % self.name

        return "Item( '%s'\n%s\n)" % (
               self.name, self._indent(options, '      '))

#-------------------------------------------------------------------------
#  'UItem' class:
#-------------------------------------------------------------------------


class UItem(Item):
    """ An Item that has no label.
    """

    show_label = Bool(False)

#-------------------------------------------------------------------------
#  'Custom' class:
#-------------------------------------------------------------------------


class Custom(Item):
    """ An Item using a 'custom' style.
    """

    style = EditorStyle('custom')

#-------------------------------------------------------------------------
#  'UCustom' class:
#-------------------------------------------------------------------------


class UCustom(Custom):
    """ An Item using a 'custom' style with no label.
    """

    show_label = Bool(False)

#-------------------------------------------------------------------------
#  'Readonly' class:
#-------------------------------------------------------------------------


class Readonly(Item):
    """ An Item using a 'readonly' style.
    """

    style = EditorStyle('readonly')

#-------------------------------------------------------------------------
#  'UReadonly' class:
#-------------------------------------------------------------------------


class UReadonly(Readonly):
    """ An Item using a 'readonly' style with no label.
    """

    show_label = Bool(False)

#-------------------------------------------------------------------------
#  'Label' class:
#-------------------------------------------------------------------------


class Label(Item):
    """ An item that is a label.
    """

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, label, item_theme=None, **traits):
        super(Label, self).__init__(
            label=label,
            item_theme=convert_theme(item_theme),
            **traits
        )

#-------------------------------------------------------------------------
#  'Heading' class:
#-------------------------------------------------------------------------


class Heading(Label):
    """ An item that is a fancy label.
    """

    # Override the 'style' trait to default to the fancy 'custom' style:
    style = Constant('custom')

#-------------------------------------------------------------------------
#  'Spring' class:
#-------------------------------------------------------------------------


class Spring(Item):
    """ An item that is a layout "spring".
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Name of the trait the item is editing
    name = 'spring'

    # Should a label be displayed?
    show_label = Bool(False)

    # Editor to use for the item
    editor = Instance('traitsui.api.NullEditor', ())

    # Should the item use extra space along its Group's layout orientation?
    springy = True

# A pre-defined spring for convenience
spring = Spring()



try:
    from json import dumps, loads
except ImportError:
    from simplejson import dumps, loads
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO


def encode(header='', payload=''):
    def textsize(s):
        if s:
            return str(len(s))
        return ''    
    return '%i,%i:%s%s' % (textsize(header), textsize(payload), header, payload)


class FileEncoder(object):
    
    def __init__(self, f):
        self.f = f
        
    def write(self, header='', payload=''):
        fwrite = self.f.write
        def textsize(s):
            if s:
                return str(len(s))
            return ''        
        fwrite('%s,%s:' % (textsize(header), textsize(payload)))
        if header:
            fwrite(header)
        if payload:
            fwrite(payload)


class JSONFileEncoder(FileEncoder):
    
    def write(self, header=None, payload=''):
        if header is None:
            super(JSONFileEncoder, self).write('', payload)
        else:         
            header_json = dumps(header, separators=(',', ':'))
            super(JSONFileEncoder, self).write(header_json, payload)


class DecoderError(Exception):
    pass

class PreludeError(DecoderError):
    pass

class Decoder(object):
    
    STAGE_PRELUDE, STAGE_SIZE, STAGE_HEADER, STAGE_PAYLOAD = range(4)
    MAX_PRELUDE = 255
        
    def __init__(self, no_prelude=False, prelude_callback=None):
        
        self.prelude_callback = prelude_callback
        self.stream_broken = False
        self.expecting_bytes = None
        self.stage = self.STAGE_PRELUDE
        self._prelude = []
        self._size = []
        self._expecting_bytes = None
        
        self.header_size = None
        self.payload_size = None
        
        self._header_bytes = None
        self._payload_bytes = None
        
        self._header_data = []
        self._payload_data = []
        
        self.header = None
        self.payload = None
        
        if no_prelude:
            self.stage = self.STAGE_SIZE
    
    
    def feed(self, data):
        
        if self.stream_broken:
            raise DecoderError('Stream is broken')
        
        STAGE_PRELUDE, STAGE_SIZE, STAGE_HEADER, STAGE_PAYLOAD = range(4)
                
        size_append = self._size.append
        header_append = self._header_data.append
        payload_append = self._payload_data.append
        datafind = data.find
        
        def reset_packet():
            self.expecting_bytes = None
            del self._header_data[:]
            del self._payload_data[:]
            self.header = None
            self.payload = None
            
        data_len = len(data)
        data_pos = 0
        expecting_bytes = self.expecting_bytes        
        stage = self.stage
        
        if stage == STAGE_PRELUDE:
            max_find = min(len(data), data_pos + self.MAX_PRELUDE)           
            cr_pos = datafind('\n', data_pos, max_find)
            if cr_pos == -1:
                self._prelude.append(data[data_pos:])
                data_pos = max_find
                if sum(len(s) for s in self._prelude) > self.MAX_PRELUDE:
                    self.stream_broken = True
                    raise PreludeError('Prelude not found')
            else:
                self._prelude.append(data[data_pos:cr_pos])
                if sum(len(s) for s in self._prelude) > self.MAX_PRELUDE:
                    self.stream_broken = True
                    raise PreludeError('Prelude not found')
                data_pos = cr_pos + 1                                
                prelude = ''.join(self._prelude)
                del self._prelude[:]
                reset_packet()
                if not self.on_prelude(prelude):
                    self.broken = True
                    return
                stage = STAGE_SIZE        
        
        while data_pos < data_len:
            
            if stage == STAGE_HEADER:
                bytes_to_read = min(data_len - data_pos, expecting_bytes)
                header_append(data[data_pos:data_pos + bytes_to_read])
                data_pos += bytes_to_read
                expecting_bytes -= bytes_to_read 
                if not expecting_bytes:                    
                    self.header = ''.join(self._header_data)                                    
                    if not self.payload_size:
                        yield self.header, ''
                        reset_packet()
                        expecting_bytes = None
                        stage = STAGE_SIZE
                    else:                        
                        stage = STAGE_PAYLOAD
                        expecting_bytes = self.payload_size
                                        
            elif stage == STAGE_PAYLOAD:
                bytes_to_read = min(data_len - data_pos, expecting_bytes)                
                payload_append(data[data_pos:data_pos + bytes_to_read])
                data_pos += bytes_to_read
                expecting_bytes -= bytes_to_read 
                if not expecting_bytes:                    
                    self.payload = ''.join(self._payload_data)
                    yield self.header, self.payload
                    reset_packet()
                    stage = STAGE_SIZE
                    expecting_bytes = None
                
            elif stage == STAGE_SIZE:
                term_pos = datafind(':', data_pos)
                if term_pos == -1:
                    size_append(data[data_pos:])                    
                    break
                else:
                    size_append(data[data_pos:term_pos])
                    data_pos = term_pos + 1                    

                size = ''.join(self._size)
                del self._size[:]
                if ',' in size:
                    header_size, payload_size = size.split(',', 1)
                else:
                    header_size = size
                    payload_size = ''                    
                try:
                    self.header_size = int(header_size or '0')
                    self.payload_size = int(payload_size or '0')
                except ValueError:
                    self.stream_broken = False
                    raise DecoderError('Invalid size in packet (%s)' % size)
                
                if self.header_size:
                    expecting_bytes = self.header_size                        
                    stage = STAGE_HEADER
                elif self.payload_size:
                    expecting_bytes = self.payload_size                        
                    stage = STAGE_PAYLOAD
                else:
                    # A completely empty packet, permitted, if a little odd
                    yield '', ''
                    reset_packet()                        
                    expecting_bytes = None

        self.expecting_bytes = expecting_bytes                
        self.stage = stage
                
                        
    def on_prelude(self, prelude):
        if self.prelude_callback and not self.prelude_callback(self, prelude):
            return False
        #pass
        #print "Prelude:", prelude
        return True


class JSONDecoder(Decoder):
    
    def feed(self, data):
        for header, payload in Decoder.feed(self, data):
            if header:
                header = loads(header)
            else:
                header = {}
            yield header, payload

   
if __name__ == "__main__":
    
    f = StringIO()
    encoder = JSONFileEncoder(f)
    encoder.write(dict(a=1, b=2), 'Payload')
    encoder.write(dict(foo="bar", nested=dict(apples="oranges"), alist=range(5)), 'Payload goes here')
    encoder.write(None, 'Payload')
    encoder.write(dict(a=1))
    encoder.write()
    
    stream = 'prelude\n' + f.getvalue()
    
    #print stream
    
#    packets = ['Prelude string\n',
#                encode('header', 'payload'),
#               encode('header number 2', 'second payload'),
#               encode('', '')]
#    
#    stream = ''.join(packets)
    
    decoder = JSONDecoder()
    
    stream = 'pyfs/0.1\n59,13:{"type":"rpc","method":"ping","client_ref":"-1221142848:1"}Hello, World!'
    
    fdata = StringIO(stream)
    
    while 1:
        data = fdata.read(3)
        if not data:
            break
        for header, payload in decoder.feed(data):
            print "Header:", repr(header)
            print "Payload:", repr(payload)   
        


from svg_parser_utils import *
from vector_math import *
import shader
import svg_shader_constants
import math


class GradientShaders:

    def __init__(self):
        self._radial_shader = None
        self._linear_shader = None

    @property
    def radial_shader(self):
        if not self._radial_shader:
            self._radial_shader = shader.make_program_from_src(
                                    "vs", "radial_ps",
                                    svg_shader_constants.vertex, svg_shader_constants.radial)
        return self._radial_shader

    @property
    def linear_shader(self):
        if not self._linear_shader:
            self._linear_shader = shader.make_program_from_src(
                                    "vs", "linear_ps",
                                    svg_shader_constants.vertex, svg_shader_constants.linear)
        return self._linear_shader

gradient_shaders = GradientShaders()


class GradientContainer(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.callback_dict = {}

    def call_me_on_add(self, callback, grad_id):
        '''
        The client wants to know when the gradient with id grad_id gets
        added.  So store this callback for when that happens.
        When the desired gradient is added, the callback will be called
        with the gradient as the first and only argument.
        '''
        cblist = self.callback_dict.get(grad_id, None)
        if cblist == None:
            cblist = [callback]
            self.callback_dict[grad_id] = cblist
            return
        cblist.append(callback)

    def update(self, *args, **kwargs):
        raise NotImplementedError('update not done for GradientContainer')

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)
        callbacks = self.callback_dict.get(key, [])
        for callback in callbacks:
            callback(val)
        
    
class Gradient(object):
    def __init__(self, element, svg):
        self.element = element
        self.stops = {}
        for e in element.getiterator():
            if e.tag.endswith('stop'):
                style = parse_style(e.get('style', ''))
                color = parse_color(e.get('stop-color'))
                if 'stop-color' in style:
                    color = parse_color(style['stop-color'])
                color[3] = int(float(e.get('stop-opacity', '1')) * 255)
                if 'stop-opacity' in style:
                    color[3] = int(float(style['stop-opacity']) * 255)
                offset = parse_float(e.get('offset'))
                self.stops[offset] = color
        self.stops = sorted(self.stops.items())
        self.svg = svg
        self.grad_transform = Matrix(element.get('gradientTransform'))
        self.inv_transform = Matrix(element.get('gradientTransform')).inverse()
        self.opacity = element.get('opacity', 1.0)
        self.units = element.get('gradientUnits', 'objectBoundingBox')
        inherit = self.element.get('{http://www.w3.org/1999/xlink}href')
        parent = None
        delay_params = False
        if inherit:
            parent_id = inherit[1:]
            parent = self.svg._gradients.get(parent_id, None)
            if parent == None:
                self.svg._gradients.call_me_on_add(self.tardy_gradient_parsed, parent_id)
                delay_params = True
                return
        if not delay_params:
            self.get_params(parent)

    def sample(self, pt, path):
        if not self.stops: return [255, 0, 255, 255]
        t = self.grad_value(self.inv_transform(pt), path)
        if t < self.stops[0][0]:
            return self.stops[0][1]
        for n, top in enumerate(self.stops[1:]):
            bottom = self.stops[n]
            if t <= top[0]:
                u = bottom[0]
                v = top[0]
                alpha = (t - u)/(v - u)
                return [int(x[0] * (1 - alpha) + x[1] * alpha) for x in zip(bottom[1], top[1])]
        return self.stops[-1][1]

    def get_params(self, parent):
        for param in self.params:
            v = None
            if parent:
                v = getattr(parent, param, None)
            my_v = self.element.get(param)
            if my_v:
                v = str(my_v)
            if v:
                setattr(self, param, v)

    def tardy_gradient_parsed(self, gradient):
        self.get_params(gradient)
        
    def apply_shader(self, path, transform, opacity):
        pass
    
    def unapply_shader(self):
        pass


class LinearGradient(Gradient):
    params = ['x1', 'x2', 'y1', 'y2', 'stops']

    def __init__(self, *args):
        self.x1 = '0'
        self.x2 = "100%"
        self.y1 = '0'
        self.y2 = '0'
        Gradient.__init__(self, *args)

    def grad_value(self, pt, path):
        return ((pt[0] - self.get_x1(path)) * (self.get_x2(path) - self.get_x1(path)) + (pt[1] - self.get_y1(path)) * (self.get_y2(path) - self.get_y1(path))) \
            / ((self.get_x1(path) - self.get_x2(path)) ** 2 + (self.get_y1(path) - self.get_y2(path)) ** 2)

    def get_x1(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.x1)
            return percentage * (maxx - minx) + minx
        else:
            return float(self.x1)

    def get_x2(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.x2)
            return percentage * (maxx - minx) + minx
        else:
            return float(self.x2)

    def get_y2(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.y1)
            return percentage * (maxy - miny) + miny
        else:
            return float(self.y1)

    def get_y1(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.y2)
            return percentage * (maxy - miny) + miny
        else:
            return float(self.y2)

    def apply_shader(self, path, transform, opacity):
        if not self.stops: return
        gradient_shaders.linear_shader.use()
        gradient_shaders.linear_shader.uniformf("opacity", self.opacity * opacity)
        gradient_shaders.linear_shader.uniformf("start", self.get_x1(path), self.get_y1(path))
        gradient_shaders.linear_shader.uniformf("end", self.get_x2(path), self.get_y2(path))
        gradient_shaders.linear_shader.uniform_matrixf("worldTransform", False, svg_matrix_to_gl_matrix(transform))
        gradient_shaders.linear_shader.uniform_matrixf("gradientTransform",
                                     False,
                                     svg_matrix_to_gl_matrix(self.grad_transform))
        gradient_shaders.linear_shader.uniform_matrixf("invGradientTransform",
                                     False,
                                     svg_matrix_to_gl_matrix(self.inv_transform))
        stop_points = []
        for stop in self.stops:
            stop_point, color = stop
            stop_points.append(stop_point)
        while len(stop_points) < 5:
            stop_points.append(0.0)

        if len(stop_points) > 5:
            stop_points = stop_points[:5]

        gradient_shaders.linear_shader.uniformf("stops", *(stop_points[1:]))
        
        def get_stop(i):
            return self.stops[i] if i < len(self.stops) else self.stops[-1]
        
        for i in xrange(5):
            stop_point, color = get_stop(i)
            color = tuple(float(x)/255.0 for x in color)
            gradient_shaders.linear_shader.uniformf("stop" + str(i), *color)
    
    def unapply_shader(self):
        if not self.stops: return
        gradient_shaders.linear_shader.stop()


class RadialGradient(Gradient):
    params = ['cx', 'cy', 'r', 'stops']

    def __init__(self, *args):
        element = args[0]
        self.cx = '50%'
        self.cy = "50%"
        self.r = "100%"
        self.fx = element.get('fx', None)
        self.fy = element.get('fy', None)
        
        
        Gradient.__init__(self, *args)

    def grad_value(self, pt, path):
        return math.sqrt((pt[0] - self.get_cx(path)) ** 2 + (pt[1] - self.get_cy(path)) ** 2) / self.get_r(path)

    def get_cx(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.cx)
            return percentage * (maxx - minx) + minx
        else: #userSpaceOnUse
            return float(self.cx)

    def get_cy(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.cy)
            return percentage * (maxy - miny) + miny
        else: #userSpaceOnUse
            return float(self.cy)

    def get_fx(self, path):
        if self.fx == None:
            return self.get_cx(path)
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.fx)
            return percentage * (maxx - minx) + minx
        else: #userSpaceOnUse
            return float(self.fx)

    def get_fy(self, path):
        if self.fy == None:
            return self.get_cy(path)
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.fy)
            return percentage * (maxy - miny) + miny
        else: #userSpaceOnUse
            return float(self.fy)

    def get_r(self, path):
        if self.units == 'objectBoundingBox':
            minx, miny, maxx, maxy = path.bounding_box()
            percentage = parse_float(self.r)
            extent = min(maxx-minx, maxy-miny)
            return percentage * extent
        else: #userSpaceOnUse
            return float(self.r)

    def apply_shader(self, path, transform, opacity):
        if not self.stops: return
        gradient_shaders.radial_shader.use()
        gradient_shaders.radial_shader.uniformf("opacity", self.opacity*opacity)
        gradient_shaders.radial_shader.uniformf("radius", self.get_r(path))
        gradient_shaders.radial_shader.uniformf("center", self.get_cx(path), self.get_cy(path))
        gradient_shaders.radial_shader.uniformf("focalPoint", self.get_fx(path), self.get_fy(path))
        gradient_shaders.radial_shader.uniform_matrixf("worldTransform", False, svg_matrix_to_gl_matrix(transform))
        gradient_shaders.radial_shader.uniform_matrixf("gradientTransform",
                                     False,
                                     svg_matrix_to_gl_matrix(self.grad_transform))
        gradient_shaders.radial_shader.uniform_matrixf("invGradientTransform",
                                     False,
                                     svg_matrix_to_gl_matrix(self.inv_transform))
        stop_points = []
        for stop in self.stops:
            stop_point, color = stop
            stop_points.append(stop_point)
        
        while len(stop_points) < 5:
            stop_points.append(0.0)
        
        #can't support more than 4 of these bad boys..
        if len(stop_points) > 5:
            stop_points = stop_points[:5]

        gradient_shaders.radial_shader.uniformf("stops", *(stop_points[1:]))
        
        def get_stop(i):
            return self.stops[i] if i < len(self.stops) else (1.0, [0.0, 0.0, 0.0, 0.0])
        
        for i in xrange(len(stop_points)):
            stop_point, color = get_stop(i)
            color = tuple(float(x)/255.0 for x in color)
            gradient_shaders.radial_shader.uniformf("stop" + str(i), *color)
    
    def unapply_shader(self):
        if not self.stops: return
        gradient_shaders.radial_shader.stop()



"""
fs.osfs.watch
=============

Change watcher support for OSFS

"""

import os
import sys
import errno
import threading

from fs.errors import *
from fs.path import *
from fs.watch import *

OSFSWatchMixin = None

#  Try using native implementation on win32
if sys.platform == "win32":
    try:
        from fs.osfs.watch_win32 import OSFSWatchMixin
    except ImportError:
        pass

#  Try using pyinotify if available
if OSFSWatchMixin is None:
    try:
        from fs.osfs.watch_inotify import OSFSWatchMixin
    except ImportError:
        pass

#  Fall back to raising UnsupportedError
if OSFSWatchMixin is None:
    class OSFSWatchMixin(object):
        def __init__(self, *args, **kwargs):
            super(OSFSWatchMixin, self).__init__(*args, **kwargs)
        def add_watcher(self,*args,**kwds):
            raise UnsupportedError
        def del_watcher(self,watcher_or_callback):
            raise UnsupportedError





#-------------------------------------------------------------------------------
#
#  Defines a horizontal or vertical flow layout sizer for wxPython
#
#  Written by: David C. Morrill
#
#  Date: 01/12/2006
#
#  (c) Copyright 2006 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from pyface.timer.api \
    import do_later

#-------------------------------------------------------------------------------
#  'FlowSizer' class:
#-------------------------------------------------------------------------------

class FlowSizer ( wx.PySizer ):

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, orient = wx.HORIZONTAL ):
        super( FlowSizer, self ).__init__()
        self._orient       = orient
        self._frozen       = False
        self._needed_size  = None

    #---------------------------------------------------------------------------
    #  Calculates the minimum size needed by the sizer:
    #---------------------------------------------------------------------------

    def CalcMin ( self ):
        """ Calculates the minimum size needed by the sizer.
        """
        if self._needed_size is not None:
            return self._needed_size

        horizontal  = (self._orient == wx.HORIZONTAL)
        dx = dy = 0

        for item in self.GetChildren() :
            idx, idy = item.CalcMin()
            if horizontal:
                dy = max( dy, idy )
            else:
                dx = max( dx, idx )

        return wx.Size( dx, dy )

    #---------------------------------------------------------------------------
    #  Layout the contents of the sizer based on the sizer's current size and
    #  position:
    #---------------------------------------------------------------------------

    def RecalcSizes ( self ):
        """ Layout the contents of the sizer based on the sizer's current size
            and position.
        """
        horizontal = (self._orient == wx.HORIZONTAL)
        x,   y     = self.GetPosition()
        dx, dy     = self.GetSize()
        x0, y0     = x, y
        ex         = x + dx
        ey         = y + dy
        mdx = mdy  = sdx = sdy = 0

        visible = True
        cur_max = 0

        for item in self.GetChildren() :
            idx, idy  = item.CalcMin()
            expand    = item.GetFlag() & wx.EXPAND
            if horizontal:
                if (x > x0) and ((x + idx) > ex):
                    x   = x0
                    y  += (mdy + sdy)
                    mdy = sdy = 0
                    if y >= ey:
                        visible = False

                cur_max = max( idy, cur_max )
                if expand:
                    idy = cur_max

                if item.IsSpacer():
                    sdy = max( sdy, idy )
                    if x == x0:
                        idx = 0
                item.SetDimension( wx.Point( x, y ), wx.Size( idx, idy ) )
                item.Show( visible )
                x  += idx
                mdy = max( mdy, idy )
            else:
                if (y > y0) and ((y + idy) > ey):
                    y   = y0
                    x  += (mdx + sdx)
                    mdx = sdx = 0
                    if x >= ex:
                        visible = False

                cur_max = max( idx, cur_max )
                if expand:
                    idx = cur_max

                if item.IsSpacer():
                    sdx = max( sdx, idx )
                    if y == y0:
                        idy = 0

                item.SetDimension( wx.Point( x, y ), wx.Size( idx, idy ) )
                item.Show( visible )
                y  += idy
                mdx = max( mdx, idx )

        if (not visible) and (self._needed_size is None):
            max_dx = max_dy = 0
            if horizontal:
                max_dy = max( dy, y + mdy + sdy - y0 )
            else:
                max_dx = max( dx, x + mdx + sdx - x0 )
            self._needed_size = wx.Size( max_dx, max_dy )
            if not self._frozen:
                self._do_parent( '_freeze' )
            do_later( self._do_parent, '_thaw' )
        else:
            self._needed_size = None

    #---------------------------------------------------------------------------
    #  Prevents the specified window from doing any further screen updates:
    #---------------------------------------------------------------------------

    def _freeze ( self, window ):
        """ Prevents the specified window from doing any further screen updates.
        """
        window.Freeze()
        self._frozen = True

    #---------------------------------------------------------------------------
    #  Lays out a specified window and then allows it to be updated again:
    #---------------------------------------------------------------------------

    def _thaw ( self, window ):
        """ Lays out a specified window and then allows it to be updated again.
        """
        window.Layout()
        window.Refresh()
        if self._frozen:
            self._frozen = False
            window.Thaw()

    #---------------------------------------------------------------------------
    #  Does a specified operation on the sizer's parent window:
    #---------------------------------------------------------------------------

    def _do_parent ( self, method ):
        """ Does a specified operation on the sizer's parent window.
        """
        i = 0
        while True:
            try:
                item = self.GetItem( i )
                if item is None:
                    break
                i += 1
            except:
                return

            if item.IsWindow():
                getattr( self, method )( item.GetWindow().GetParent() )
                return

